
from epyk.core.css import Colors
from epyk.core.css import Defaults_css


def autoPrefixer(prop):
  """
  Description:
  ------------
  CSS Style function to return the different attributes names for the compatibility with the main browsers.
  This function is used everytime there is a need for a CSS extension.

  The main browsers supported are IE, Chrome, Firefox and Opera

  Related Pages:
  --------------
  https://www.w3schools.com/cssref/css3_browsersupport.asp

  :param prop: String. The CSS Attribute key
  :return: A generator function with the different keys to be added to the style
  """
  map = {
    'animation': ['-webkit-', '-moz-', '-o-'],
    'animation-direction': ['-webkit-', '-moz-', '-o-'],
    'animation-duration': ['-webkit-', '-moz-', '-o-'],
    'animation-iteration-count': ['-webkit-', '-moz-', '-o-'],
    'animation-timing-function': ['-webkit-', '-moz-', '-o-'],
    'animation-name': ['-webkit-', '-moz-', '-o-'],
    'appearance': ['-webkit-', '-moz-'],
    'backface-visibility': ['-webkit-'],
    'box-decoration-break': ['-webkit-'],
    'font-kerning': ['-webkit-'],
    'font-variant-ligatures': ['-webkit-'],
    'hyphens': ['-webkit-'],
    'image-rendering': ['-moz-'],
    'mask': ['-webkit-'],
    'mask-type': ['-webkit-'],
    'tab-size': ['-moz-'],
    'text-combine-upright': ['-webkit-'],
    'text-decoration-color': ['-webkit-'],
    'text-decoration-line': ['-webkit-'],
    'text-emphasis': ['-webkit-'],
    'text-emphasis-color': ['-webkit-'],
    'text-emphasis-position': ['-webkit-'],
    'text-emphasis-style': ['-webkit-'],
    'text-orientation': ['-webkit-'],
    'transition': ['-webkit-', '-moz-', '-o-'],
    'transition-property': ['-webkit-', '-moz-', '-o-'],
    'transition-duration': ['-webkit-', '-moz-', '-o-'],
    'transition-timing-function': ['-webkit-', '-moz-', '-o-'],
    'transition-delay': ['-webkit-', '-moz-', '-o-'],
    'user-select': ['-webkit-', '-moz-', '-o-', '-khtml-', '-ms-'],
      }
  for pref in map.get(prop, []):
    yield "%s%s" % (pref, prop)


class CssMixin(object):

  @property
  def align_content(self): return self.htmlObj.css("align-content")

  @align_content.setter
  def align_content(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"align-content": val})

  @property
  def align_items(self): return self.htmlObj.css("align-items")

  @align_items.setter
  def align_items(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"align-items": val})

  @property
  def align_self(self): return self.htmlObj.css("align-self")

  @align_self.setter
  def align_self(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"align-self": val})

  @property
  def all(self): return self.htmlObj.css("all")

  @all.setter
  def all(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"all": val})

  @property
  def animation(self):
    """
    Description:
    ------------
    CSS allows animation of HTML elements without using JavaScript or Flash!

    Related Pages:

			https://www.w3schools.com/css/css3_animations.asp
    """
    return self.htmlObj.css("animation")

  @animation.setter
  def animation(self, val):
    val = val if val is not None else 'None'
    for m_val in autoPrefixer("animation"):
      self.htmlObj.css({m_val: val})
    self.htmlObj.css({"animation": val})

  @property
  def animation_delay(self): return self.htmlObj.css("animation-delay")

  @animation_delay.setter
  def animation_delay(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"animation-delay": val})

  @property
  def animation_direction(self):
    """
    Description:
    ------------
    The animation-direction property defines whether an animation should be played forwards, backwards or in alternate cycles.

    Related Pages:

			https://www.w3schools.com/cssref/css3_pr_animation-direction.asp
    """
    return self.htmlObj.css("animation-direction")

  @animation_direction.setter
  def animation_direction(self, val):
    val = val if val is not None else 'None'
    for m_val in autoPrefixer("animation-direction"):
      self.htmlObj.css({m_val: val})
    self.htmlObj.css({"animation-direction": val})

  @property
  def animation_duration(self):
    """
    Description:
    ------------
    The animation-duration property defines how long an animation should take to complete one cycle.

    Related Pages:

			https://www.w3schools.com/cssref/css3_pr_animation-duration.asp
    """
    return self.htmlObj.css("animation-duration")

  @animation_duration.setter
  def animation_duration(self, val):
    val = val if val is not None else 'None'
    for m_val in autoPrefixer("animation-duration"):
      self.htmlObj.css({m_val: val})
    self.htmlObj.css({"animation-duration": val})

  @property
  def animation_fill_mode(self):
    """
    Description:
    ------------
    The animation-fill-mode property specifies a style for the target element when the animation is not playing (before it starts, after it ends, or both)
    The animation-fill-mode property can have the following values:

    none - Default value. Animation will not apply any styles to the element before or after it is executing
    forwards - The element will retain the style values that is set by the last keyframe (depends on animation-direction and animation-iteration-count)
    backwards - The element will get the style values that is set by the first keyframe (depends on animation-direction), and retain this during the animation-delay period
    both - The animation will follow the rules for both forwards and backwards, extending the animation properties in both directions

    Related Pages:

			https://www.w3schools.com/cssref/css3_pr_animation-fill-mode.asp
    """
    return self.htmlObj.css("animation-fill-mode")

  @animation_fill_mode.setter
  def animation_fill_mode(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"animation-fill-mode": val})

  @property
  def animation_iteration_count(self):
    """
    Description:
    ------------
    The animation-iteration-count property specifies the number of times an animation should be played.

    Related Pages:

			https://www.w3schools.com/cssref/css3_pr_animation-iteration-count.asp
    """
    return self.htmlObj.css("animation-iteration-count")

  @animation_iteration_count.setter
  def animation_iteration_count(self, val):
    val = val if val is not None else 'None'
    for m_val in autoPrefixer("animation-iteration-count"):
      self.htmlObj.css({m_val: val})
    self.htmlObj.css({"animation-iteration-count": val})

  @property
  def animation_name(self):
    """
    Description:
    ------------
    The animation-name property specifies a name for the @keyframes animation.

    Related Pages:

			https://www.w3schools.com/cssref/css3_pr_animation-name.asp
    """
    return self.htmlObj.css("animation-name")

  @animation_name.setter
  def animation_name(self, val):
    val = val if val is not None else 'None'
    for m_val in autoPrefixer("animation-name"):
      self.htmlObj.css({m_val: val})
    self.htmlObj.css({"animation-name": val})

  @property
  def animation_play_state(self): return self.htmlObj.css("animation-play-state")

  @animation_play_state.setter
  def animation_play_state(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"animation-play-state": val})

  @property
  def animation_timing_function(self):
    """
    Description:
    ------------
    The animation-timing-function specifies the speed curve of an animation.

    Related Pages:

			https://www.w3schools.com/cssref/css3_pr_animation-timing-function.asp
    """
    return self.htmlObj.css("animation-timing-function")

  @animation_timing_function.setter
  def animation_timing_function(self, val):
    val = val if val is not None else 'None'
    for m_val in autoPrefixer("animation-timing-function"):
      self.htmlObj.css({m_val: val})
    self.htmlObj.css({"animation-timing-function": val})

  @property
  def appearance(self):
    """
    Description:
    ------------
    This property is frequently used in XUL stylesheets to design custom widgets with platform-appropriate styling.
    It is also used in the XBL implementations of the widgets that ship with the Mozilla platform.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/CSS/appearance
    """
    return self.htmlObj.css(autoPrefixer("appearance")[0])

  @appearance.setter
  def appearance(self, val):
    val = val if val is not None else 'None'
    for m_val in autoPrefixer("appearance"):
      self.htmlObj.css({m_val: val})

  @property
  def backface_visibility(self):
    """
    Description:
    ------------
    Hide and show the back face of two rotated <div> elements:

    Related Pages:

			https://www.w3schools.com/cssref/css3_pr_backface-visibility.asp
    """
    return self.htmlObj.css("backface-visibility")

  @backface_visibility.setter
  def backface_visibility(self, val):
    val = val if val is not None else 'None'
    for m_val in autoPrefixer("backface-visibility"):
      self.htmlObj.css({m_val: val})
    self.htmlObj.css({"backface-visibility": val})

  @property
  def background(self):
    """
    Description:
    ------------
    Set different background properties in one declaration

    Related Pages:

			https://www.w3schools.com/cssref/css3_pr_background.asp
    """
    return self.htmlObj.css("background")

  @background.setter
  def background(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"background": val})

  @property
  def background_attachment(self): return self.htmlObj.css("background-attachment")

  @background_attachment.setter
  def background_attachment(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"background-attachment": val})

  @property
  def background_blend_mode(self): return self.htmlObj.css("background-blend-mode")

  @background_blend_mode.setter
  def background_blend_mode(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"background-blend-mode": val})

  @property
  def background_clip(self): return self.htmlObj.css("background-clip")

  @background_clip.setter
  def background_clip(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"background-clip": val})

  @property
  def background_color(self):
    """
    Description:
    ------------
    The background-color property sets the background color of an element.

    The background of an element is the total size of the element, including padding and border (but not the margin).

    Related Pages:

			https://www.w3schools.com/cssref/pr_background-color.asp
    """
    return self.htmlObj.css("background-color")

  @background_color.setter
  def background_color(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"background-color": val})

  @property
  def background_image(self): return self.htmlObj.css("background-image")

  @background_image.setter
  def background_image(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"background-image": val})

  @property
  def background_origin(self): return self.htmlObj.css("background-origin")

  @background_origin.setter
  def background_origin(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"background-origin": val})

  @property
  def background_position(self): return self.htmlObj.css("background-position")

  @background_position.setter
  def background_position(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"background-position": val})

  @property
  def background_repeat(self):
    """
    Description:
    ------------
    The background-repeat property sets if/how a background image will be repeated.

    Usage::

      https://www.w3schools.com/cssref/pr_background-repeat.asp
    """
    return self.htmlObj.css("background-repeat")

  @background_repeat.setter
  def background_repeat(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"background-repeat": val})

  @property
  def background_size(self): return self.htmlObj.css("background-size")

  @background_size.setter
  def background_size(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"background-size": val})

  @property
  def border(self):
    """
    Description:
    ------------
    The border property is a shorthand property for:
    - border-width
    - border-style (required)
    - border-color

    Related Pages:

			https://www.w3schools.com/cssref/pr_border.asp

    Usage::

      https://www.w3schools.com/cssref/pr_border.asp
    """
    return self.htmlObj.css("border")

  @border.setter
  def border(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"border": val})

  @property
  def border_bottom(self): return self.htmlObj.css("border-bottom")

  @border_bottom.setter
  def border_bottom(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"border-bottom": val})

  @property
  def border_bottom_color(self): return self.htmlObj.css("border-bottom-color")

  @border_bottom_color.setter
  def border_bottom_color(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"border-bottom-color": val})

  @property
  def border_bottom_left_radius(self): return self.htmlObj.css("border-bottom-left-radius")

  @border_bottom_left_radius.setter
  def border_bottom_left_radius(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"border-bottom-left-radius": val})

  @property
  def border_bottom_right_radius(self): return self.htmlObj.css("border-bottom-right-radius")

  @border_bottom_right_radius.setter
  def border_bottom_right_radius(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"border-bottom-right-radius": val})

  @property
  def border_bottom_style(self): return self.htmlObj.css("border-bottom-style")

  @border_bottom_style.setter
  def border_bottom_style(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"border-bottom-style": val})

  @property
  def border_bottom_width(self): return self.htmlObj.css("border-bottom-width")

  @border_bottom_width.setter
  def border_bottom_width(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"border-bottom-width": val})

  @property
  def border_collapse(self):
    """
    Description:
    ------------
    The border-collapse property sets whether table borders should collapse into a single border or be separated as in standard HTML.

    Related Pages:

			https://www.w3schools.com/cssref/pr_border-collapse.asp
    """
    return self.htmlObj.css("border-collapse")

  @border_collapse.setter
  def border_collapse(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"border-collapse": val})

  @property
  def border_color(self): return self.htmlObj.css("border-color")

  @border_color.setter
  def border_color(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"border-color": val})

  @property
  def border_image(self): return self.htmlObj.css("border-image")

  @border_image.setter
  def border_image(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"border-image": val})

  @property
  def border_image_outset(self): return self.htmlObj.css("border-image-outset")

  @border_image_outset.setter
  def border_image_outset(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"border-image-outset": val})

  @property
  def border_image_repeat(self): return self.htmlObj.css("border-image-repeat")

  @border_image_repeat.setter
  def border_image_repeat(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"border-image-repeat": val})

  @property
  def border_image_slice(self): return self.htmlObj.css("border-image-slice")

  @border_image_slice.setter
  def border_image_slice(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"border-image-slice": val})

  @property
  def border_image_source(self): return self.htmlObj.css("border-image-source")

  @border_image_source.setter
  def border_image_source(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"border-image-source": val})

  @property
  def border_image_width(self): return self.htmlObj.css("border-image-width")

  @border_image_width.setter
  def border_image_width(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"border-image-width": val})

  @property
  def border_left(self): return self.htmlObj.css("border-left")

  @border_left.setter
  def border_left(self, val):
    val = val if val is not None else 'None'
    if isinstance(val, int):
      self.htmlObj.css({"border-left": "%spx" % val})
    else:
      self.htmlObj.css({"border-left": val})

  @property
  def border_left_color(self): return self.htmlObj.css("border-left-color")

  @border_left_color.setter
  def border_left_color(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"border-left-color": val})

  @property
  def border_left_style(self): return self.htmlObj.css("border-left-style")

  @border_left_style.setter
  def border_left_style(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"border-left-style": val})

  @property
  def border_left_width(self): return self.htmlObj.css("border-left-width")

  @border_left_width.setter
  def border_left_width(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"border-left-width": val})

  @property
  def border_radius(self):
    """
    Description:
    ------------
    The border-radius property defines the radius of the element's corners.

    Related Pages:

			https://www.w3schools.com/cssref/css3_pr_border-radius.asp
    """
    return self.htmlObj.css("border-radius")

  @border_radius.setter
  def border_radius(self, val):
    val = val if val is not None else 'None'
    if isinstance(val, int):
      val = "%spx" % val
    self.htmlObj.css({"border-radius": val})

  @property
  def border_right(self): return self.htmlObj.css("border-right")

  @border_right.setter
  def border_right(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"border-right": val})

  @property
  def border_right_color(self): return self.htmlObj.css("border-right-color")

  @border_right_color.setter
  def border_right_color(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"border-right-color": val})

  @property
  def border_right_style(self): return self.htmlObj.css("border-right-style")

  @border_right_style.setter
  def border_right_style(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"border-right-style": val})

  @property
  def border_right_width(self): return self.htmlObj.css("border-right-width")

  @border_right_width.setter
  def border_right_width(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"border-right-width": val})

  @property
  def border_spacing(self):
    """
    Description:
    ------------
    The border-spacing property sets the distance between the borders of adjacent cells.

    Related Pages:

			https://www.w3schools.com/cssref/pr_border-spacing.asp
    """
    return self.htmlObj.css("border-spacing")

  @border_spacing.setter
  def border_spacing(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"border-spacing": val})

  @property
  def border_style(self): return self.htmlObj.css("border-style")

  @border_style.setter
  def border_style(self, val):
    val = val if val is not None else 'None'
    if isinstance(val, int):
      val = "%spx" % val
    self.htmlObj.css({"border-style": val})

  @property
  def border_top(self): return self.htmlObj.css("border-top")

  @border_top.setter
  def border_top(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"border-top": val})

  @property
  def border_top_color(self): return self.htmlObj.css("border-top-color")

  @border_top_color.setter
  def border_top_color(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"border-top-color": val})

  @property
  def border_top_left_radius(self): return self.htmlObj.css("border-top-left-radius")

  @border_top_left_radius.setter
  def border_top_left_radius(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"border-top-left-radius": val})

  @property
  def border_top_right_radius(self): return self.htmlObj.css("border-top-right-radius")

  @border_top_right_radius.setter
  def border_top_right_radius(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"border-top-right-radius": val})

  @property
  def border_top_style(self): return self.htmlObj.css("border-top-style")

  @border_top_style.setter
  def border_top_style(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"border-top-style": val})

  @property
  def border_top_width(self): return self.htmlObj.css("border-top-width")

  @border_top_width.setter
  def border_top_width(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"border-top-width": val})

  @property
  def border_width(self): return self.htmlObj.css("border-width")

  @border_width.setter
  def border_width(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"border-width": val})

  @property
  def bottom(self):
    """
    Description:
    ------------
    The bottom property affects the vertical position of a positioned element. This property has no effect on non-positioned elements.

    Related Pages:

			https://www.w3schools.com/cssref/pr_pos_bottom.asp
    """
    return self.htmlObj.css("bottom")

  @bottom.setter
  def bottom(self, val):
    val = val if val is not None else 'None'
    if isinstance(val, int):
      val = "%spx" % val
    self.htmlObj.css({"bottom": val})

  @property
  def box_decoration_break(self):
    """
    Description:
    ------------
    The box-decoration-break property specifies how the background, padding, border, border-image, box-shadow, margin, and clip-path of an element is applied when the box for the element is fragmented.

    Related Pages:

			https://www.w3schools.com/cssref/css3_pr_box-decoration-break.asp
    """
    return self.htmlObj.css("box-decoration-break")

  @box_decoration_break.setter
  def box_decoration_break(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"box-decoration-break": val})

  @property
  def box_shadow(self):
    """
    Description:
    ------------
    The box-shadow property attaches one or more shadows to an element.

    Related Pages:

			https://www.w3schools.com/cssref/css3_pr_box-shadow.asp
    """
    return self.htmlObj.css("box-shadow")

  @box_shadow.setter
  def box_shadow(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"box-shadow": val})

  @property
  def box_sizing(self): return self.htmlObj.css("box-sizing")

  @box_sizing.setter
  def box_sizing(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"box-sizing": val})

  @property
  def caption_side(self): return self.htmlObj.css("caption-side")

  @caption_side.setter
  def caption_side(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"caption-side": val})

  @property
  def caret_color(self): return self.htmlObj.css("caret-color")

  @caret_color.setter
  def caret_color(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"caret-color": val})

  @property
  def clear(self): return self.htmlObj.css("clear")

  @clear.setter
  def clear(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"clear": val})

  @property
  def clip(self): return self.htmlObj.css("clip")

  @clip.setter
  def clip(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"clip": val})

  @property
  def clip_path(self): return self.htmlObj.css("clip-path")

  @clip_path.setter
  def clip_path(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"clip-path": val})

  @property
  def color(self):
    """
    Description:
    ------------
    The color property specifies the color of text.

    Tip: Use a background color combined with a text color that makes the text easy to read.

    Related Pages:

			https://www.w3schools.com/cssref/pr_text_color.asp
    """
    return self.htmlObj.css("color")

  @color.setter
  def color(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"color": val})

  @property
  def column_count(self): return self.htmlObj.css("column-count")

  @column_count.setter
  def column_count(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"column-count": val})

  @property
  def column_fill(self): return self.htmlObj.css("column-fill")

  @column_fill.setter
  def column_fill(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"column-fill": val})

  @property
  def column_gap(self): return self.htmlObj.css("column-gap")

  @column_gap.setter
  def column_gap(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"column-gap": val})

  @property
  def column_rule(self): return self.htmlObj.css("column-rule")

  @column_rule.setter
  def column_rule(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"column-rule": val})

  @property
  def column_rule_color(self): return self.htmlObj.css("column-rule-color")

  @column_rule_color.setter
  def column_rule_color(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"column-rule-color": val})

  @property
  def column_rule_style(self): return self.htmlObj.css("column-rule-style")

  @column_rule_style.setter
  def column_rule_style(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"column-rule-style": val})

  @property
  def column_rule_width(self): return self.htmlObj.css("column-rule-width")

  @column_rule_width.setter
  def column_rule_width(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"column-rule-width": val})

  @property
  def column_span(self): return self.htmlObj.css("column-span")

  @column_span.setter
  def column_span(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"column-span": val})

  @property
  def column_width(self): return self.htmlObj.css("column-width")

  @column_width.setter
  def column_width(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"column-width": val})

  @property
  def columns(self): return self.htmlObj.css("columns")

  @columns.setter
  def columns(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"columns": val})

  @property
  def content(self): return self.htmlObj.css("content")

  @content.setter
  def content(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"content": val})

  @property
  def counter_increment(self): return self.htmlObj.css("counter-increment")

  @counter_increment.setter
  def counter_increment(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"counter-increment": val})

  @property
  def counter_reset(self): return self.htmlObj.css("counter-reset")

  @counter_reset.setter
  def counter_reset(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"counter-reset": val})

  @property
  def cursor(self):
    """
    Description:
    ------------
    The cursor property specifies the mouse cursor to be displayed when pointing over an element.

    Related Pages:

			https://www.w3schools.com/cssref/pr_class_cursor.asp
    """
    return self.htmlObj.css("cursor")

  @cursor.setter
  def cursor(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"cursor": val})

  @property
  def direction(self): return self.htmlObj.css("direction")

  @direction.setter
  def direction(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"direction": val})

  @property
  def display(self):
    """
    Description:
    ------------
    The display property specifies the display behavior (the type of rendering box) of an element.

    In HTML, the default display property value is taken from the HTML specifications or from the browser/user default style sheet. The default value in XML is inline, including SVG elements.

    Related Pages:

			https://www.w3schools.com/cssref/pr_class_display.asp
    """
    return self.htmlObj.css("display")

  @display.setter
  def display(self, val):
    val = val if val is not None else 'None'
    if val is False:
      val = 'None'
    self.htmlObj.css({"display": val})

  @property
  def empty_cells(self): return self.htmlObj.css("empty-cells")

  @empty_cells.setter
  def empty_cells(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"empty-cells": val})

  @property
  def filter(self): return self.htmlObj.css("filter")

  @filter.setter
  def filter(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"filter": val})

  @property
  def flex(self): return self.htmlObj.css("flex")

  @flex.setter
  def flex(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"flex": val})

  @property
  def flex_basis(self): return self.htmlObj.css("flex-basis")

  @flex_basis.setter
  def flex_basis(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"flex-basis": val})

  @property
  def flex_direction(self): return self.htmlObj.css("flex-direction")

  @flex_direction.setter
  def flex_direction(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"flex-direction": val})

  @property
  def flex_flow(self): return self.htmlObj.css("flex-flow")

  @flex_flow.setter
  def flex_flow(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"flex-flow": val})

  @property
  def flex_grow(self): return self.htmlObj.css("flex-grow")

  @flex_grow.setter
  def flex_grow(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"flex-grow": val})

  @property
  def flex_shrink(self): return self.htmlObj.css("flex-shrink")

  @flex_shrink.setter
  def flex_shrink(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"flex-shrink": val})

  @property
  def flex_wrap(self): return self.htmlObj.css("flex-wrap")

  @flex_wrap.setter
  def flex_wrap(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"flex-wrap": val})

  @property
  def float(self):
    """
    Description:
    ------------
    The float property specifies how an element should float

    Related Pages:

			https://www.w3schools.com/cssref/pr_class_float.asp
    """
    return self.htmlObj.css("float")

  @float.setter
  def float(self, val):
    val = val if val is not None else 'none'
    defined_vals = set(['none', 'left', 'right', 'initial', 'inherit'])
    if Defaults_css.CSS_EXCEPTIONS and val not in defined_vals:
      raise Exception(Defaults_css.CSS_EXCEPTIONS_FORMAT % ("float", val))

    self.htmlObj.css({"float": val})

  @property
  def font(self): return self.htmlObj.css("font")

  @font.setter
  def font(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"font": val})

  @property
  def font_family(self):
    """
    Description:
    ------------
    The font-family property specifies the font for an element.

    Related Pages:

			https://www.w3schools.com/cssref/pr_font_font-family.asp
    """
    return self.htmlObj.css("font-family")

  @font_family.setter
  def font_family(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"font-family": val})

  @property
  def font_kerning(self): return self.htmlObj.css("font-kerning")

  @font_kerning.setter
  def font_kerning(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"font-kerning": val})

  @property
  def font_size(self):
    """
    Description:
    ------------
    The font-size property sets the size of a font.

    Related Pages:

			https://www.w3schools.com/cssref/pr_font_font-size.asp
    """
    return self.htmlObj.css("font-size")

  @font_size.setter
  def font_size(self, val):
    if isinstance(val, int):
      val = "%spx" % val
    val = val if val is not None else 'None'
    self.htmlObj.css({"font-size": val})

  @property
  def font_size_adjust(self): return self.htmlObj.css("font-size-adjust")

  @font_size_adjust.setter
  def font_size_adjust(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"font-size-adjust": val})

  @property
  def font_stretch(self): return self.htmlObj.css("font-stretch")

  @font_stretch.setter
  def font_stretch(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"font-stretch": val})

  @property
  def font_style(self): return self.htmlObj.css("font-style")

  @font_style.setter
  def font_style(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"font-style": val})

  @property
  def font_variant(self): return self.htmlObj.css("font-variant")

  @font_variant.setter
  def font_variant(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"font-variant": val})

  @property
  def font_weight(self):
    """
    Description:
    ------------
    The font-weight property sets how thick or thin characters in text should be displayed.

    Related Pages:

			https://www.w3schools.com/cssref/pr_font_weight.asp
    """
    return self.htmlObj.css("font-weight")

  @font_weight.setter
  def font_weight(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"font-weight": val})

  @property
  def grid(self): return self.htmlObj.css("grid")

  @grid.setter
  def grid(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"grid": val})

  @property
  def grid_area(self): return self.htmlObj.css("grid-area")

  @grid_area.setter
  def grid_area(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"grid-area": val})

  @property
  def grid_auto_columns(self): return self.htmlObj.css("grid-auto-columns")

  @grid_auto_columns.setter
  def grid_auto_columns(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"grid-auto-columns": val})

  @property
  def grid_auto_flow(self): return self.htmlObj.css("grid-auto-flow")

  @grid_auto_flow.setter
  def grid_auto_flow(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"grid-auto-flow": val})

  @property
  def grid_auto_rows(self): return self.htmlObj.css("grid-auto-rows")

  @grid_auto_rows.setter
  def grid_auto_rows(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"grid-auto-rows": val})

  @property
  def grid_column(self): return self.htmlObj.css("grid-column")

  @grid_column.setter
  def grid_column(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"grid-column": val})

  @property
  def grid_column_end(self): return self.htmlObj.css("grid-column-end")

  @grid_column_end.setter
  def grid_column_end(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"grid-column-end": val})

  @property
  def grid_column_gap(self): return self.htmlObj.css("grid-column-gap")

  @grid_column_gap.setter
  def grid_column_gap(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"grid-column-gap": val})

  @property
  def grid_column_start(self): return self.htmlObj.css("grid-column-start")

  @grid_column_start.setter
  def grid_column_start(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"grid-column-start": val})

  @property
  def grid_gap(self): return self.htmlObj.css("grid-gap")

  @grid_gap.setter
  def grid_gap(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"grid-gap": val})

  @property
  def grid_row(self): return self.htmlObj.css("grid-row")

  @grid_row.setter
  def grid_row(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"grid-row": val})

  @property
  def grid_row_end(self): return self.htmlObj.css("grid-row-end")

  @grid_row_end.setter
  def grid_row_end(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"grid-row-end": val})

  @property
  def grid_row_gap(self): return self.htmlObj.css("grid-row-gap")

  @grid_row_gap.setter
  def grid_row_gap(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"grid-row-gap": val})

  @property
  def grid_row_start(self): return self.htmlObj.css("grid-row-start")

  @grid_row_start.setter
  def grid_row_start(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"grid-row-start": val})

  @property
  def grid_template(self): return self.htmlObj.css("grid-template")

  @grid_template.setter
  def grid_template(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"grid-template": val})

  @property
  def grid_template_areas(self): return self.htmlObj.css("grid-template-areas")

  @grid_template_areas.setter
  def grid_template_areas(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"grid-template-areas": val})

  @property
  def grid_template_columns(self): return self.htmlObj.css("grid-template-columns")

  @grid_template_columns.setter
  def grid_template_columns(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"grid-template-columns": val})

  @property
  def grid_template_rows(self): return self.htmlObj.css("grid-template-rows")

  @grid_template_rows.setter
  def grid_template_rows(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"grid-template-rows": val})

  @property
  def hanging_punctuation(self): return self.htmlObj.css("hanging-punctuation")

  @hanging_punctuation.setter
  def hanging_punctuation(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"hanging-punctuation": val})

  @property
  def height(self):
    """
    Description:
    ------------
    The height property sets the height of an element.

    The height of an element does not include padding, borders, or margins!

    Related Pages:

			https://www.w3schools.com/cssref/pr_dim_height.asp
    """
    return self.htmlObj.css("height")

  @height.setter
  def height(self, val):
    val = val if val is not None else 'None'
    if isinstance(val, int):
      val = "%spx" % val
    self.htmlObj.css({"height": val})

  @property
  def hyphens(self): return self.htmlObj.css("hyphens")

  @hyphens.setter
  def hyphens(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"hyphens": val})

  @property
  def isolation(self): return self.htmlObj.css("isolation")

  @isolation.setter
  def isolation(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"isolation": val})

  @property
  def justify_content(self):
    """
    Description:
    ------------
    The justify-content property aligns the flexible container's items when the items do not use all available space on the main-axis (horizontally).
    This property is not inherited

    Related Pages:

			https://www.w3schools.com/cssref/css3_pr_justify-content.asp
    """
    return self.htmlObj.css("justify-content")

  @justify_content.setter
  def justify_content(self, val):
    vals = set(["flex-start", "flex-end", "center", "space-between", "space-around", "initial", "inherit"])
    val = val if val is not None else 'None'
    self.htmlObj.css({"justify-content": val})

  @property
  def left(self):
    """
    Description:
    ------------
    The left property affects the horizontal position of a positioned element. This property has no effect on non-positioned elements.

    Related Pages:

			https://www.w3schools.com/cssref/pr_pos_left.asp
    """
    return self.htmlObj.css("left")

  @left.setter
  def left(self, val):
    if isinstance(val, int):
      val = "%spx" % val
    val = val if val is not None else 'None'
    self.htmlObj.css({"left": val})

  @property
  def letter_spacing(self): return self.htmlObj.css("letter-spacing")

  @letter_spacing.setter
  def letter_spacing(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"letter-spacing": val})

  @property
  def line_height(self):
    """
    Description:
    ------------
    The line-height property specifies the height of a line.

    Note: Negative values are not allowed.

    Related Pages:

			https://www.w3schools.com/cssref/pr_dim_line-height.asp
    """
    return self.htmlObj.css("line-height")

  @line_height.setter
  def line_height(self, val):
    if isinstance(val, int):
      val = "%spx" % val
    val = val if val is not None else 'None'
    self.htmlObj.css({"line-height": val})

  @property
  def list_style(self): return self.htmlObj.css("list-style")

  @list_style.setter
  def list_style(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"list-style": val})

  @property
  def list_style_image(self): return self.htmlObj.css("list-style-image")

  @list_style_image.setter
  def list_style_image(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"list-style-image": val})

  @property
  def list_style_position(self): return self.htmlObj.css("list-style-position")

  @list_style_position.setter
  def list_style_position(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"list-style-position": val})

  @property
  def list_style_type(self): return self.htmlObj.css("list-style-type")

  @list_style_type.setter
  def list_style_type(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"list-style-type": val})

  @property
  def margin(self):
    """
    Description:
    ------------
    The margin property sets the margins for an element, and is a shorthand property for the following properties:
    - margin-top
    - margin-right
    - margin-bottom
    - margin-left

    Related Pages:

			https://www.w3schools.com/cssref/pr_margin.asp
    """
    return self.htmlObj.css("margin")

  @margin.setter
  def margin(self, val):
    if isinstance(val, int):
      val = "%spx" % val
    val = val if val is not None else 'None'
    self.htmlObj.css({"margin": val})

  @property
  def margin_bottom(self):
    """
    Description:
    ------------
    The margin-bottom property sets the bottom margin of an element.

    Note: Negative values are allowed.

    Related Pages:

			https://www.w3schools.com/cssref/pr_margin-bottom.asp
    """
    return self.htmlObj.css("margin-bottom")

  @margin_bottom.setter
  def margin_bottom(self, val):
    val = val if val is not None else 'None'
    if isinstance(val, int):
      val = "%spx" % val
    self.htmlObj.css({"margin-bottom": val})

  @property
  def margin_left(self):
    """
    Description:
    ------------
    The margin-left property sets the left margin of an element.

    Note: Negative values are allowed.

    Related Pages:

			https://www.w3schools.com/cssref/pr_margin-left.asp
    """
    return self.htmlObj.css("margin-left")

  @margin_left.setter
  def margin_left(self, val):
    if isinstance(val, int):
      val = "%spx" % val
    val = val if val is not None else 'None'
    self.htmlObj.css({"margin-left": val})

  @property
  def margin_right(self):
    """
    Description:
    ------------
    The margin-right property sets the right margin of an element.

    Note: Negative values are allowed.

    Related Pages:

			https://www.w3schools.com/cssref/pr_margin-right.asp
    """
    return self.htmlObj.css("margin-right")

  @margin_right.setter
  def margin_right(self, val):
    val = val if val is not None else 'None'
    if isinstance(val, int):
      val = "%spx" % val
    self.htmlObj.css({"margin-right": val})

  @property
  def margin_top(self):
    """
    Description:
    ------------
    The margin-top property sets the top margin of an element.

    Note: Negative values are allowed.

    Related Pages:

			https://www.w3schools.com/cssref/pr_margin-top.asp
    """
    return self.htmlObj.css("margin-top")

  @margin_top.setter
  def margin_top(self, val):
    if isinstance(val, int):
      val = "%spx" % val
    val = val if val is not None else 'None'
    self.htmlObj.css({"margin-top": val})

  @property
  def max_height(self): return self.htmlObj.css("max-height")

  @max_height.setter
  def max_height(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"max-height": val})

  @property
  def max_width(self): return self.htmlObj.css("max-width")

  @max_width.setter
  def max_width(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"max-width": val})

  @property
  def min_height(self):
    """
    Description:
    ------------
    The min-height property defines the minimum height of an element.

    Related Pages:

			https://www.w3schools.com/cssref/pr_dim_min-height.asp
    """
    return self.htmlObj.css("min-height")

  @min_height.setter
  def min_height(self, val):
    val = val if val is not None else 'None'
    if isinstance(val, int):
      val = "%spx" % val
    self.htmlObj.css({"min-height": val})

  @property
  def min_width(self): return self.htmlObj.css("min-width")

  @min_width.setter
  def min_width(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"min-width": val})

  @property
  def mix_blend_mode(self): return self.htmlObj.css("mix-blend-mode")

  @mix_blend_mode.setter
  def mix_blend_mode(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"mix-blend-mode": val})

  @property
  def object_fit(self): return self.htmlObj.css("object-fit")

  @object_fit.setter
  def object_fit(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"object-fit": val})

  @property
  def object_position(self): return self.htmlObj.css("object-position")

  @object_position.setter
  def object_position(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"object-position": val})

  @property
  def opacity(self):
    """
    Description:
    ------------
    The opacity property sets the opacity level for an element.

    The opacity-level describes the transparency-level, where 1 is not transparent at all, 0.5 is 50% see-through, and 0 is completely transparent.

    Related Pages:

			https://www.w3schools.com/cssref/css3_pr_opacity.asp
    """
    return self.htmlObj.css("opacity")

  @opacity.setter
  def opacity(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"opacity": val})

  @property
  def order(self): return self.htmlObj.css("order")

  @order.setter
  def order(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"order": val})

  @property
  def outline(self): return self.htmlObj.css("outline")

  @outline.setter
  def outline(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"outline": val})

  @property
  def outline_color(self): return self.htmlObj.css("outline-color")

  @outline_color.setter
  def outline_color(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"outline-color": val})

  @property
  def outline_offset(self): return self.htmlObj.css("outline-offset")

  @outline_offset.setter
  def outline_offset(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"outline-offset": val})

  @property
  def outline_style(self): return self.htmlObj.css("outline-style")

  @outline_style.setter
  def outline_style(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"outline-style": val})

  @property
  def outline_width(self): return self.htmlObj.css("outline-width")

  @outline_width.setter
  def outline_width(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"outline-width": val})

  @property
  def overflow(self):
    """
    Description:
    ------------
    The overflow property specifies what should happen if content overflows an element's box.

    This property specifies whether to clip content or to add scrollbars when an element's content is too big to fit in a specified area.

    Note: The overflow property only works for block elements with a specified height.

    Values: visible|hidden|scroll|auto|initial|inherit

    Related Pages:

			https://www.w3schools.com/cssref/pr_pos_overflow.asp
    """
    return self.htmlObj.css("overflow")

  @overflow.setter
  def overflow(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"overflow": val})

  @property
  def overflow_x(self):
    """
    Description:
    ------------
    The overflow-x property specifies whether to clip the content, add a scroll bar, or display overflow content of a block-level element, when it overflows at the left and right edges.

    Related Pages:

			https://www.w3schools.com/cssref/css3_pr_overflow-x.asp
    """
    return self.htmlObj.css("overflow-x")

  @overflow_x.setter
  def overflow_x(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"overflow-x": val})

  @property
  def overflow_y(self):
    """
    Description:
    ------------
    The overflow-y property specifies whether to clip the content, add a scroll bar, or display overflow content of a block-level element, when it overflows at the top and bottom edges.

    Related Pages:

			https://www.w3schools.com/cssref/css3_pr_overflow-y.asp
    """
    return self.htmlObj.css("overflow-y")

  @overflow_y.setter
  def overflow_y(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"overflow-y": val})

  @property
  def padding(self):
    """
    Description:
    ------------
    The CSS padding properties are used to generate space around an element's content, inside of any defined borders.

    Related Pages:

			https://www.w3schools.com/css/css_padding.asp
    """
    return self.htmlObj.css("padding")

  @padding.setter
  def padding(self, val):
    val = val if val is not None else 'None'
    if isinstance(val, int):
      val = "%spx" % val
    self.htmlObj.css({"padding": val})

  @property
  def padding_bottom(self): return self.htmlObj.css("padding-bottom")

  @padding_bottom.setter
  def padding_bottom(self, val):
    val = val if val is not None else 'None'
    if isinstance(val, int):
      val = "%spx" % val
    self.htmlObj.css({"padding-bottom": val})

  @property
  def padding_left(self):
    """
    Description:
    ------------
    An element's padding is the space between its content and its border.

    The padding-left property sets the left padding (space) of an element.

    Related Pages:

			https://www.w3schools.com/cssref/pr_padding-left.asp
    """
    return self.htmlObj.css("padding-left")

  @padding_left.setter
  def padding_left(self, val):
    val = val if val is not None else 'None'
    if isinstance(val, int):
      val = "%spx" % val
    self.htmlObj.css({"padding-left": val})

  @property
  def padding_right(self):
    """
    Description:
    ------------
    An element's padding is the space between its content and its border

    Related Pages:

			https://www.w3schools.com/cssref/pr_padding-right.asp
    """
    return self.htmlObj.css("padding-right")

  @padding_right.setter
  def padding_right(self, val):
    val = val if val is not None else 'None'
    if isinstance(val, int):
      val = "%spx" % val
    self.htmlObj.css({"padding-right": val})

  @property
  def padding_top(self):
    """
    Description:
    ------------
    An element's padding is the space between its content and its border.

    The padding-top property sets the top padding (space) of an element.

    Related Pages:

			https://www.w3schools.com/cssref/pr_padding-top.asp
    """
    return self.htmlObj.css("padding-top")

  @padding_top.setter
  def padding_top(self, val):
    if isinstance(val, int):
      val = "%spx" % val
    val = val if val is not None else 'None'
    self.htmlObj.css({"padding-top": val})

  @property
  def page_break_after(self): return self.htmlObj.css("page-break-after")

  @page_break_after.setter
  def page_break_after(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"page-break-after": val})

  @property
  def page_break_before(self): return self.htmlObj.css("page-break-before")

  @page_break_before.setter
  def page_break_before(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"page-break-before": val})

  @property
  def page_break_inside(self): return self.htmlObj.css("page-break-inside")

  @page_break_inside.setter
  def page_break_inside(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"page-break-inside": val})

  @property
  def perspective(self): return self.htmlObj.css("perspective")

  @perspective.setter
  def perspective(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"perspective": val})

  @property
  def perspective_origin(self): return self.htmlObj.css("perspective-origin")

  @perspective_origin.setter
  def perspective_origin(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"perspective-origin": val})

  @property
  def pointer_events(self): return self.htmlObj.css("pointer-events")

  @pointer_events.setter
  def pointer_events(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"pointer-events": val})

  @property
  def position(self):
    """
    Description:
    ------------
    The position property specifies the type of positioning method used for an element (static, relative, absolute, fixed, or sticky).

    Related Pages:

			https://www.w3schools.com/cssref/pr_class_position.asp
    """
    return self.htmlObj.css("position")

  @position.setter
  def position(self, val):
    val = val if val is not None else 'None'
    if val in ["sticky"]:
      self.htmlObj.css({"position": "-webkit-%s" % val})
    self.htmlObj.css({"position": val})

  @property
  def quotes(self): return self.htmlObj.css("quotes")

  @quotes.setter
  def quotes(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"quotes": val})

  @property
  def resize(self): return self.htmlObj.css("resize")

  @resize.setter
  def resize(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"resize": val})

  @property
  def right(self):
    """
    Description:
    ------------
    The right property affects the horizontal position of a positioned element.
    This property has no effect on non-positioned elements.

    Related Pages:

			https://www.w3schools.com/cssref/pr_pos_right.asp
    """
    return self.htmlObj.css("right")

  @right.setter
  def right(self, val):
    val = val if val is not None else 'None'
    if isinstance(val , int):
      val = "%spx" % val
    self.htmlObj.css({"right": val})

  @property
  def scroll_behavior(self): return self.htmlObj.css("scroll-behavior")

  @scroll_behavior.setter
  def scroll_behavior(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"scroll-behavior": val})

  @property
  def tab_size(self): return self.htmlObj.css("tab-size")

  @tab_size.setter
  def tab_size(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"tab-size": val})

  @property
  def table_layout(self): return self.htmlObj.css("table-layout")

  @table_layout.setter
  def table_layout(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"table-layout": val})

  @property
  def text_align(self):
    """
    Description:
    ------------
    The text-align property specifies the horizontal alignment of text in an element.

    Related Pages:

			https://www.w3schools.com/cssref/pr_text_text-align.ASP
    """
    return self.htmlObj.css("text-align")

  @text_align.setter
  def text_align(self, val):
    val = val if val is not None else 'None'
    if Defaults_css.CSS_EXCEPTIONS and val not in ['center', 'None', 'left', 'right', 'justify', 'inherit', 'initial']:
      raise Exception(Defaults_css.CSS_EXCEPTIONS_FORMAT % ("text_align", val))

    self.htmlObj.css({"text-align": val})

  @property
  def text_align_last(self): return self.htmlObj.css("text-align-last")

  @text_align_last.setter
  def text_align_last(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"text-align-last": val})

  @property
  def text_decoration(self): return self.htmlObj.css("text-decoration")

  @text_decoration.setter
  def text_decoration(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"text-decoration": val})

  @property
  def text_decoration_color(self): return self.htmlObj.css("text-decoration-color")

  @text_decoration_color.setter
  def text_decoration_color(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"text-decoration-color": val})

  @property
  def text_decoration_line(self): return self.htmlObj.css("text-decoration-line")

  @text_decoration_line.setter
  def text_decoration_line(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"text-decoration-line": val})

  @property
  def text_decoration_style(self): return self.htmlObj.css("text-decoration-style")

  @text_decoration_style.setter
  def text_decoration_style(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"text-decoration-style": val})

  @property
  def text_indent(self): return self.htmlObj.css("text-indent")

  @text_indent.setter
  def text_indent(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"text-indent": val})

  @property
  def text_justify(self): return self.htmlObj.css("text-justify")

  @text_justify.setter
  def text_justify(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"text-justify": val})

  @property
  def text_overflow(self): return self.htmlObj.css("text-overflow")

  @text_overflow.setter
  def text_overflow(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"text-overflow": val})

  @property
  def text_shadow(self):
    """
    Description:
    ------------
    The text-shadow property adds shadow to text.

    This property accepts a comma-separated list of shadows to be applied to the text.

    Related Pages:

			https://www.w3schools.com/cssref/css3_pr_text-shadow.asp
    """
    return self.htmlObj.css("text-shadow")

  @text_shadow.setter
  def text_shadow(self, val):
    val = val if val is not None else 'None'
    if isinstance(val, list):
      val = ",".join(val)
    self.htmlObj.css({"text-shadow": val})

  @property
  def text_transform(self):
    """
    Description:
    ------------
    The text-transform property controls the capitalization of text.

    Related Pages:

			https://www.w3schools.com/cssref/pr_text_text-transform.asp
    """
    return self.htmlObj.css("text-transform")

  @text_transform.setter
  def text_transform(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"text-transform": val})

  @property
  def text_stoke(self):
    """
    Description:
    ------------
    The -webkit-text-stroke CSS property specifies the width and color of strokes for text characters. This is a shorthand property for the longhand properties -webkit-text-stroke-width and -webkit-text-stroke-color.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/CSS/-webkit-text-stroke
    """
    return self.htmlObj.css("-webkit-text-stroke")

  @text_stoke.setter
  def text_stoke(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"-webkit-text-stroke": val})

  @property
  def text_stoke_width(self):
    """
    Description:
    ------------
    The -webkit-text-stroke-width CSS property specifies the width of the stroke for text.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/CSS/-webkit-text-stroke-width
    """
    return self.htmlObj.css("-webkit-text-stroke-width")

  @text_stoke_width.setter
  def text_stoke_width(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"-webkit-text-stroke-width": val})

  @property
  def text_stoke_color(self):
    """
    Description:
    ------------
    The -webkit-text-stroke-color CSS property specifies the stroke color of characters of text. If this property is not set, the value of the color property is used.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/CSS/-webkit-text-stroke-color
    """
    return self.htmlObj.css("-webkit-text-stroke-color")

  @text_stoke_color.setter
  def text_stoke_color(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"-webkit-text-stroke-color": val})

  @property
  def top(self):
    """
    Description:
    ------------
    The top property affects the vertical position of a positioned element.
    This property has no effect on non-positioned elements.

    Related Pages:

			https://www.w3schools.com/cssref/pr_pos_top.asp
    """
    return self.htmlObj.css("top")

  @top.setter
  def top(self, val):
    if val is None:
      val = 'None'
    if isinstance(val, int):
      val = "%spx" % val
    self.htmlObj.css({"top": val})

  @property
  def transform(self):
    """
    Description:
    ------------
    The transform property applies a 2D or 3D transformation to an element. This property allows you to rotate, scale, move, skew, etc., elements.

    Related Pages:

			https://www.w3schools.com/cssref/css3_pr_transform.asp
    """
    return self.htmlObj.css("transform")

  @transform.setter
  def transform(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"transform": val})

  @property
  def transform_origin(self): return self.htmlObj.css("transform-origin")

  @transform_origin.setter
  def transform_origin(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"transform-origin": val})

  @property
  def transform_style(self): return self.htmlObj.css("transform-style")

  @transform_style.setter
  def transform_style(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"transform-style": val})

  @property
  def transition(self): return self.htmlObj.css("transition")

  @transition.setter
  def transition(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"transition": val})

  @property
  def transition_delay(self): return self.htmlObj.css("transition-delay")

  @transition_delay.setter
  def transition_delay(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"transition-delay": val})

  @property
  def transition_duration(self): return self.htmlObj.css("transition-duration")

  @transition_duration.setter
  def transition_duration(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"transition-duration": val})

  @property
  def transition_property(self): return self.htmlObj.css("transition-property")

  @transition_property.setter
  def transition_property(self, val):
    val = val if val is not None else 'None'
    for m_val in autoPrefixer("transition-property"):
      self.htmlObj.css({m_val: val})
    self.htmlObj.css({"transition-property": val})

  @property
  def transition_timing_function(self):
    return self.htmlObj.css("transition-timing-function")

  @transition_timing_function.setter
  def transition_timing_function(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"transition-timing-function": val})

  @property
  def unicode_bidi(self): return self.htmlObj.css("unicode-bidi")

  @unicode_bidi.setter
  def unicode_bidi(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"unicode-bidi": val})

  @property
  def user_select(self):
    """
    Description:
    ------------
    The user-select property specifies whether the text of an element can be selected.

    In web browsers, if you double-click on some text it will be selected/highlighted. This property can be used to prevent this.

    Related Pages:

			https://www.w3schools.com/cssref/css3_pr_user-select.asp
    """
    return self.htmlObj.css("user-select")

  @user_select.setter
  def user_select(self, val):
    val = val if val is not None else 'None'
    for m_val in autoPrefixer("user-select"):
      self.htmlObj.css({m_val: val})
    self.htmlObj.css({"user-select": val})

  @property
  def vertical_align(self):
    """
    Description:
    ------------
    The vertical-align property sets the vertical alignment of an element

    Related Pages:

			https://www.w3schools.com/cssref/pr_pos_vertical-align.asp
    """
    return self.htmlObj.css("vertical-align")

  @vertical_align.setter
  def vertical_align(self, val):
    val = val if val is not None else 'None'
    if Defaults_css.CSS_EXCEPTIONS and val not in ["baseline", "length", "sub", "super", "top", "text-top", "middle",
                                                   "bottom", "text-bottom", "initial", "inherit", 'None']:
      raise Exception(Defaults_css.CSS_EXCEPTIONS_FORMAT % ("vertical_align", val))

    self.htmlObj.css({"vertical-align": val})

  @property
  def visibility(self): return self.htmlObj.css("visibility")

  @visibility.setter
  def visibility(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"visibility": val})

  @property
  def white_space(self): return self.htmlObj.css("white-space")

  @white_space.setter
  def white_space(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"white-space": val})

  @property
  def width(self):
    """
    Description:
    ------------
    The width property sets the width of an element.

    The width of an element does not include padding, borders, or margins!

    Related Pages:

			https://www.w3schools.com/cssref/pr_dim_width.asp
    """
    return self.htmlObj.css("width")

  @width.setter
  def width(self, val):
    val = val if val is not None else 'None'
    if isinstance(val, int):
      val = "%spx" % val
    self.htmlObj.css({"width": val})

  @property
  def word_break(self): return self.htmlObj.css("word-break")

  @word_break.setter
  def word_break(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"word-break": val})

  @property
  def word_spacing(self): return self.htmlObj.css("word-spacing")

  @word_spacing.setter
  def word_spacing(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"word-spacing": val})

  @property
  def word_wrap(self): return self.htmlObj.css("word-wrap")

  @word_wrap.setter
  def word_wrap(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"word-wrap": val})

  @property
  def writing_mode(self): return self.htmlObj.css("writing-mode")

  @writing_mode.setter
  def writing_mode(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"writing-mode": val})

  @property
  def z_index(self):
    """
    Description:
    ------------
    The z-index property specifies the stack order of an element.

    An element with greater stack order is always in front of an element with a lower stack order.

    Note: z-index only works on positioned elements (position: absolute, position: relative, position: fixed, or position: sticky).

    Related Pages:

			https://www.w3schools.com/cssref/pr_pos_z-index.asp
    """
    return self.htmlObj.css("z-index")

  @z_index.setter
  def z_index(self, val):
    val = val if val is not None else 'None'
    self.htmlObj.css({"z-index": val})

  def middle(self):
    self.vertical_align = "middle"
    self.text_align = "center"
    return self

  def sticky(self, top=0, bottom=None, left=None, right=None):
    """
    Description:
    ------------
    An element with position: sticky; is positioned based on the user's scroll position

    Usage::

      htmlObj.style.css.sticky()

    Related Pages:

			https://www.w3schools.com/howto/tryit.asp?filename=tryhow_css_sticky_element

    Attributes:
    ----------
    :param top: Integer. The number of pixel from the top
    :param bottom:
    :param left:
    :param right:

    :return: The CSS object to allow the chaining
    """
    if top is not None:
      self.top = top
    elif bottom is not None:
      self.bottom = bottom
    if left is not None:
      self.left = left
    elif right is not None:
      self.right = right
    self.position = "sticky"
    return self

  def shadow(self, color=None, size=5):
    """
    Description:
    ------------
    Add a default dark shadow to the container

    Usage::

      .style.css.shadow()

    Related Pages:

			https://www.w3schools.com/css/css3_shadows.asp

    Attributes:
    ----------
    :param color: String. The color
    :param size: Number. The size for the shadow
    """
    if color is None:
      color = self.orign_htmlObj._report.theme.colors[2]
    self.css({"box-shadow": "%(size)spx %(size)spx %(size)spx %(color)s" % {'color': color, 'size': size}})
    return self

  def shadow_box(self, hexa_color=None, opacity=0.2, size=2, position='right'):
    """
    Description:
    ------------
    Set the box shadow color

    Related Pages:

			https://www.w3schools.com/css/css3_shadows.asp
			https://gist.github.com/ocean90/1268328

    Attributes:
    ----------
    :param hexa_color: String. An hexadecimal color code
    :param opacity:

    :return: The CSS object to allow the functions chaining
    """
    rgb = Colors.getHexToRgb(self.orign_htmlObj._report.theme.greys[-1] if hexa_color is None else hexa_color)
    #self.box_shadow = "0 2px 4px 0 rgba(%(r)s, %(g)s, %(b)s, %(opac)s), 0 3px 10px 0 rgba(%(r)s, %(g)s, %(b)s, %(opac)s)" % {"r": rgb[0], "g": rgb[1], "b": rgb[2], 'opac': opacity}
    if position == 'right':
      self.box_shadow = "%(size)spx 0 %(size)spx -%(size)spx rgba(%(r)s, %(g)s, %(b)s, %(opac)s)" % {"r": rgb[0], "g": rgb[1], "b": rgb[2], 'opac': opacity, 'size': size}
    elif position == 'left':
      self.box_shadow = "-%(size)spx 0 5px -%(size)spx rgba(%(r)s, %(g)s, %(b)s, %(opac)s)" % {"r": rgb[0], "g": rgb[1], "b": rgb[2], 'opac': opacity, 'size': size}
    else:
      self.box_shadow = "0 0 %(size)spx rgba(%(r)s, %(g)s, %(b)s, %(opac)s)" % {"r": rgb[0], "g": rgb[1], "b": rgb[2], 'opac': opacity, 'size': size}
    return self

  def shadow_text(self):
    """
    Description:
    ------------

    Related Pages:

			https://www.w3schools.com/css/css3_shadows.asp
    """
    self.text_shadow = "0 0 3px #FF0000, 0 0 5px #0000FF"
    return self

  def fixed_top(self, top=5):
    """
    Description:
    ------------
    Fix the component on the top of the page

    Usage::

      p.style.css.fixed_top()

    Attributes:
    ----------
    :param top: Number. The margin with the page top border
    """
    self.position = "fixed"
    self.display = "block"
    self.top = top
    return self

  def fixed_bottom(self, bottom=10):
    """
    Description:
    ------------
    Fix the component at the bottom of the page

    Usage::

      p.style.css.fixed_bottom()

    Attributes:
    ----------
    :param bottom: Number. The margin with the bottom of the page
    """
    self.position = "fixed"
    self.display = "block"
    self.bottom = bottom
    return self

  def borders(self, color=None, size=1, style="solid"):
    """
    Description:
    ------------
    Shortcut function to set the border color for a defined HTML component

    Usage::

      a.style.css.borders()

    Attributes:
    ----------
    :param color: Optional. The border color. Default the grey from selected theme
    :param size: Optional. The border size. Default 1pz
    :param style: Optional. The border style. Default solid - a plain line
    """
    if color is None:
      color = self.orign_htmlObj._report.theme.greys[-1]
    self.border = "%spx %s %s" % (size, style, color)
    return self

  def borders_light(self, color=None, size=1, style="solid"):
    """
    Description:
    ------------
    Shortcut function to set the light border color for a defined HTML component

    Usage::

      a.style.css.borders_light()

    Attributes:
    ----------
    :param color: Optional. The border color. Default the grey from selected theme
    :param size: Optional. The border size. Default 1pz
    :param style: Optional. The border style. Default solid - a plain line
    """
    if color is None:
      color = self.orign_htmlObj._report.theme.greys[3]
    return self.borders(color, size, style)

  def bold(self):
    """
    Description:
    ------------
    Shotcut to put the text in bold
    """
    self.font_weight = "bold"
    return self

  def underline(self):
    """
    Description:
    ------------
    Shotcut to put the underline the text

    Related Pages:

			https://www.w3schools.com/cssref/pr_text_text-decoration.asp
    """
    self.text_decoration = "underline"
    return self




from epyk.core import html


class Drawers(object):

  def __init__(self, context):
    self.context = context

  def drawer(self, width=(100, '%'), height=(100, '%'), options=None, profile=None, helper=None):
    """
    Description:
    ------------
    Bespoke drawer with handle on the right.

    Attributes:
    ----------
    :param width: Optional. A tuple with the integer for the component width and its unit
    :param height: Optional. A tuple with the integer for the component height and its unit
    :param options: Optional. A dictionary with the components properties
    :param profile: Optional. A flag to set the component performance storage
    :param helper:
    """
    h_drawer = html.HtmlDrawer.Drawer(self.context.rptObj, width, height, options, helper, profile)
    h_drawer.style.css.min_height = 200
    return h_drawer

  def left(self, width=(100, '%'), height=(200, 'px'), options=None, profile=None, helper=None):
    """
    Description:
    ------------
    Bespoke drawer with handle on the left.

    Attributes:
    ----------
    :param width: Optional. A tuple with the integer for the component width and its unit
    :param height: Optional. A tuple with the integer for the component height and its unit
    :param options:
    :param profile: Optional. A flag to set the component performance storage
    :param helper:
    """
    dflt_options = {"side": "left"}
    if options is not None:
      dflt_options.update(options)
    h_drawer = html.HtmlDrawer.Drawer(self.context.rptObj, width, height, dflt_options, helper, profile)
    return h_drawer

  def right(self, width=(100, '%'), height=(200, 'px'), options=None, profile=None, helper=None):
    """
    Description:
    ------------
    Bespoke drawer with handle on the left.

    Attributes:
    ----------
    :param width: Optional. A tuple with the integer for the component width and its unit
    :param height: Optional. A tuple with the integer for the component height and its unit
    :param options: Optional. A dictionary with the components properties
    :param profile: Optional. A flag to set the component performance storage
    :param helper:
    """
    dflt_options = {"side": "right"}
    if options is not None:
      dflt_options.update(options)
    h_drawer = html.HtmlDrawer.Drawer(self.context.rptObj, width, height, dflt_options, helper, profile)
    return h_drawer

  def multi(self, component, width=(100, '%'), height=(200, 'px'), options=None, profile=None, helper=None):
    """

    :return:
    """
    dflt_options = {"side": "right"}
    if options is not None:
      dflt_options.update(options)
    h_drawer = html.HtmlDrawer.DrawerMulti(self.context.rptObj, component, width, height, dflt_options, helper, profile)
    return h_drawer

  def no_handle(self, component, width=(100, '%'), height=(200, 'px'), options=None, profile=None, helper=None):
    """
    Description:
    ------------
    Bespoke drawer without handle.
    The event to display the panel will be attached to the component.

    Attributes:
    ----------
    :param component: Html component. Object in charge of managing the panel display
    :param width: Optional. A tuple with the integer for the component width and its unit
    :param height: Optional. A tuple with the integer for the component height and its unit
    :param options: Optional. A dictionary with the components properties
    :param profile: Optional. A flag to set the component performance storage
    :param helper:
    """
    options = options or {}
    options["side"] = 'right'
    h_drawer = html.HtmlDrawer.Drawer(self.context.rptObj, width, height, options, helper, profile)
    h_drawer.set_handle(component)
    return h_drawer





from epyk.core.html import Html
from epyk.core.html.options import OptPanel
from epyk.core.js.packages import JsQuery


class Popup(Html.Html):
  requirements = ('bootstrap', )
  name = 'Popup Container'

  def __init__(self, report, components, width, height, options, profile):
    super(Popup, self).__init__(report, [], css_attrs={"width": width, "height": height}, profile=profile)
    self.height, self.width = "%s%s" % (height[0], height[1]) if height is not None else "100%", width
    self.__options = OptPanel.OptionPopup(self, options)
    if not isinstance(components, list):
      components = [components]
    for component in components:
      self.__add__(component)

    if self.options.background:
      self.frameWidth = "%s%s" % (width[0], width[1])
      self.css({'width': '100%', 'position': 'fixed', 'height': '100%', 'background-color': 'rgba(0,0,0,0.4)', 'left': 0, 'top': 0, 'margin': 'auto'})
      self.css({'display': 'none', 'z-index': '10000', 'text-align': 'center', 'padding-top': '100px', 'padding-left': self.options.margin, 'padding-right': self.options.margin})
    else:
      self.frameWidth = "100%"
      self.css({'position': 'absolute', 'margin': 0, 'padding': 0, 'display': 'none', 'z-index': '10000'})

    if self.options.draggable:
      self.draggable(source_event=JsQuery.decorate_var(self.dom.querySelector("#%s_content" % self.htmlCode)))
    self.set_attrs(name="name", value="report_popup")
    self.keyup.escape(self.dom.hide().r, source_event="document")

  @property
  def options(self):
    """
    Property to set all the possible object for a button

    :rtype: OptPanel.OptionPopup
    """
    return self.__options

  def html(self):
    trTitle, closePopup = '', ''
    str_html = "\n".join([val.html() if hasattr(val, 'html') else str(val) for val in self.val])
    content = '''
      <table id="%(htmlCode)s_table" style="width:%(frameWidth)s;margin:auto">
        %(title)s
        <tr>
          <td style="padding:10px">
            <div style="width:100%%;text-align:right">%(closePopup)s</div>
            <div  class='scroll_content' id="%(htmlCode)s_content" style="overflow:auto;width:100%%">%(objects)s</div>
          </td>
        </tr>
      </table>''' % {'title': trTitle, 'htmlCode': self.htmlCode, 'objects': str_html, 'height': self.height, "frameWidth": self.frameWidth, 'closePopup': closePopup}
    return '''<div %s>%s</div>''' % (self.get_attrs(pyClassNames=self.style.get_classes()), content)




from epyk.core.css.styles import GrpCls
from epyk.core.css import Classes


class ClassVisTimeline(GrpCls.ClassHtml):

  def __init__(self, htmlObj):
    super(ClassVisTimeline, self).__init__(htmlObj)
    self._css_vis_items, self._css_vis_items_overflow = None, None
    self.classList['other'].add(self.css_items)
    self.classList['other'].add(self.css_items_overflow)

  @property
  def css_items(self):
    """
    """
    if self._css_vis_items is None:
      self._css_vis_items = Classes.CatalogChart.CatalogChart(self.htmlObj._report, self.classList['other'],
                                                              html_id=self.htmlObj.htmlCode).vis_items()
    return self._css_vis_items

  @property
  def css_items_overflow(self):
    """
    """
    if self._css_vis_items_overflow is None:
      self._css_vis_items_overflow = Classes.CatalogChart.CatalogChart(self.htmlObj._report, self.classList['other'],
                                                                       html_id=self.htmlObj.htmlCode).vis_items_overflow()
    return self._css_vis_items_overflow




from epyk.core.html.options import Options


class OptionsLi(Options):

  @property
  def li_css(self):
    """
    """
    return self.get({})

  @li_css.setter
  def li_css(self, css):
    self.set(css)

  @property
  def li_class(self):
    """
    """
    return self.get([])

  @li_class.setter
  def li_class(self, cls_names):
    self.set(cls_names)


class OptionsItems(Options):

  @property
  def style(self):
    """
    Description:
    ------------
    Item CSS Style
    """
    return self._config_get({})

  @style.setter
  def style(self, attrs):
    self._config(attrs)

  @property
  def badge(self):
    """
    Description:
    ------------
    Get the badge style
    """
    return self._config_get({})

  @badge.setter
  def badge(self, attrs):
    self._config(attrs)

  @property
  def delete(self):
    """
    Description:
    ------------
    Add a delete icon
    """
    return self._config_get(False)

  @delete.setter
  def delete(self, attrs):
    self._config(attrs)

  @property
  def checked(self):
    """
    Description:
    ------------
    Check default value for radio and check lists
    """
    return self._config_get(False)

  @checked.setter
  def checked(self, attrs):
    self._config(attrs)

  @property
  def icon(self):
    """
    Description:
    ------------
    Check default value for radio and check lists
    """
    return self._config_get("")

  @icon.setter
  def icon(self, attrs):
    self._config(attrs)


class OptionsTagItems(Options):

  @property
  def visible(self):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param css: Dictionary. All the CSS attributes to add the any items
    """
    return self._config_get(False)

  @visible.setter
  def visible(self, attrs):
    self._config(attrs)

  @property
  def category_css(self):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param css: Dictionary. All the CSS attributes to add the any items
    """
    return self._config_get({})

  @category_css.setter
  def category_css(self, attrs):
    self._config(attrs)

  @property
  def value_css(self):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param css: Dictionary. All the CSS attributes to add the any items
    """
    return self._config_get({})

  @value_css.setter
  def value_css(self, attrs):
    self._config(attrs)

  @property
  def item_css(self):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param css: Dictionary. All the CSS attributes to add the any items
    """
    return self._config_get({})

  @item_css.setter
  def item_css(self, attrs):
    self._config(attrs)

  @property
  def category(self):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param css: Dictionary. All the CSS attributes to add the any items
    """
    return self._config_get({})

  @category.setter
  def category(self, attrs):
    self._config(attrs)

  @property
  def icon_css(self):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param css: Dictionary. All the CSS attributes to add the any items
    """
    return self._config_get({})

  @icon_css.setter
  def icon_css(self, attrs):
    self._config(attrs)




from epyk.core.html import graph


class Plotly2D(object):
  def __init__(self, context):
    self.parent = context
    self.chartFamily = "Plotly"

  def line(self, record=None, y_columns=None, x_axis=None, profile=None, options=None, width=(100, "%"), height=(330, "px"), htmlCode=None):
    """
    Description:
    ------------

    Related Pages:

			https://plot.ly/javascript/
    https://plot.ly/javascript/plotlyjs-function-reference/#common-parameters

    Attributes:
    ----------
    :param record: List of dict. The Python recordset
    :param y_columns: List. The columns corresponding to keys in the dictionaries in the record
    :param x_axis: String. The column corresponding to a key in the dictionaries in the record
    :param profile:
    :param options:
    :param width: Tuple. The width of the component in the page, default (100, '%')
    :param height: Tuple. The height of the component in the page, default (330, "px")
    :param htmlCode:
    """
    options = options or {}
    options.update({'y_columns': y_columns, 'x_column': x_axis, 'type': 'bar', 'mode': 'lines+markers'})
    data = self.parent.context.rptObj.data.plotly.xy(record, y_columns, x_axis)
    line_chart = graph.GraphPlotly.Line(self.parent.context.rptObj, width, height, options or {}, htmlCode, profile)
    line_chart.options.responsive = True
    for d in data['datasets']:
      line_chart.add_trace(d)
    return line_chart

  def bar(self, record=None, y_columns=None, x_axis=None, profile=None, options=None, width=(100, "%"), height=(330, "px"), htmlCode=None):
    """
    Description:
    ------------

    Related Pages:

			https://plotly.com/javascript/bar-charts/

    Attributes:
    ----------
    :param record: List of dict. The Python recordset
    :param y_columns: List. The columns corresponding to keys in the dictionaries in the record
    :param x_axis: String. The column corresponding to a key in the dictionaries in the record
    :param profile:
    :param options:
    :param width: Tuple. The width of the component in the page, default (100, '%')
    :param height: Tuple. The height of the component in the page, default (330, "px")
    :param htmlCode:
    """
    options = options or {}
    options.update({'y_columns': y_columns, 'x_column': x_axis, 'type': 'bar', 'mode': None})
    data = self.parent.context.rptObj.data.plotly.xy(record, y_columns, x_axis)
    bar_chart = graph.GraphPlotly.Bar(self.parent.context.rptObj, width, height, options, htmlCode, profile)
    for d in data['datasets']:
      bar_chart.add_trace(d)
    return bar_chart

  def hbar(self, record, y_columns=None, x_axis=None, profile=None, options=None, width=(100, "%"), height=(330, "px"), htmlCode=None):
    """
    Description:
    ------------

    Related Pages:

			https://plotly.com/javascript/bar-charts/

    Attributes:
    ----------
    :param record: List of dict. The Python recordset
    :param y_columns: List. The columns corresponding to keys in the dictionaries in the record
    :param x_axis: String. The column corresponding to a key in the dictionaries in the record
    :param profile:
    :param options:
    :param width: Tuple. The width of the component in the page, default (100, '%')
    :param height: Tuple. The height of the component in the page, default (330, "px")
    :param htmlCode:
    """
    options = options or {}
    options.update({'y_columns': y_columns, 'x_column': x_axis, 'type': 'bar', 'mode': None, 'attrs': {'orientation': 'h'}})
    data = self.parent.context.rptObj.data.plotly.xy(record, y_columns, x_axis)
    bar_chart = graph.GraphPlotly.Bar(self.parent.context.rptObj, width, height, options, htmlCode, profile)
    for d in data['datasets']:
      bar_chart.add_trace(d, type='bar')
      bar_chart.data.orientation = 'h'
    return bar_chart

  def scatter(self, record=None, y_columns=None, x_axis=None, text_column=None, profile=None, options=None,
              width=(100, "%"), height=(330, "px"), htmlCode=None):
    """
    Description:
    ------------

    Related Pages:

			https://plotly.com/javascript/text-and-annotations/

    Attributes:
    ----------
    :param record: List of dict. The Python recordset
    :param y_columns: List. The columns corresponding to keys in the dictionaries in the record
    :param x_axis: String. The column corresponding to a key in the dictionaries in the record
    :param profile:
    :param options:
    :param width: Tuple. The width of the component in the page, default (100, '%')
    :param height: Tuple. The height of the component in the page, default (330, "px")
    :param htmlCode:
    """
    options = options or {}
    options.update({'y_columns': y_columns, 'x_column': x_axis, 'text_column': text_column, 'type': 'scatter',
                    'mode': 'markers+text' if text_column is not None else 'markers'})
    data = self.parent.context.rptObj.data.plotly.xy_text(record, y_columns, x_axis, text_column)
    sc_chart = graph.GraphPlotly.Line(self.parent.context.rptObj, width, height, options or {}, htmlCode, profile)
    for i, d in enumerate(data['datasets']):
      sc_chart.add_trace(d, mode=options['mode'], type=options['type'])
      sc_chart.data.marker.color = self.parent.context.rptObj.theme.colors[i]
      if text_column is not None:
        sc_chart.data.text = d['text']
    sc_chart.layout.no_background()
    return sc_chart

  def timeseries(self, record, y_columns=None, x_axis=None, profile=None, options=None, width=(100, "%"), height=(330, "px"), htmlCode=None):
    """
    Description:
    ------------

    Related Pages:

			https://plot.ly/javascript/
    https://plot.ly/javascript/plotlyjs-function-reference/#common-parameters

    Attributes:
    ----------
    :param record: List of dict. The Python recordset
    :param y_columns: List. The columns corresponding to keys in the dictionaries in the record
    :param x_axis: String. The column corresponding to a key in the dictionaries in the record
    :param profile:
    :param options:
    :param width: Tuple. The width of the component in the page, default (100, '%')
    :param height: Tuple. The height of the component in the page, default (330, "px")
    :param htmlCode:
    """
    options = options or {}
    options.update({'y_columns': y_columns, 'x_column': x_axis, 'mode': 'lines', 'type': "scatter"})
    data = self.parent.context.rptObj.data.plotly.xy(record, y_columns, x_axis)
    sc_chart = graph.GraphPlotly.Line(self.parent.context.rptObj, width, height, options, htmlCode, profile)
    for d in data['datasets']:
      sc_chart.add_trace(d)
    return sc_chart

  def scattergl(self, record, y_columns=None, x_axis=None, profile=None, options=None, width=(100, "%"), height=(330, "px"), htmlCode=None):
    """
    Description:
    ------------

    Related Pages:

			https://plotly.com/javascript/line-and-scatter/

    Attributes:
    ----------
    :param record: List of dict. The Python recordset
    :param y_columns: List. The columns corresponding to keys in the dictionaries in the record
    :param x_axis: String. The column corresponding to a key in the dictionaries in the record
    :param profile:
    :param options:
    :param width: Tuple. The width of the component in the page, default (100, '%')
    :param height: Tuple. The height of the component in the page, default (330, "px")
    :param htmlCode:
    """
    options = options or {}
    options.update({'y_columns': y_columns, 'x_column': x_axis, 'mode': 'markers', 'type': "scattergl"})
    data = self.parent.context.rptObj.data.plotly.xy(record, y_columns, x_axis)
    sc_chart = graph.GraphPlotly.Line(self.parent.context.rptObj, width, height, options or {}, htmlCode, profile)
    for d in data['datasets']:
      sc_chart.add_trace(d, type="scattergl", mode='markers')
    return sc_chart

  def histogram(self, record, y_columns=None, x_columns=None, profile=None, options=None, width=(100, "%"), height=(330, "px"), htmlCode=None):
    """

    Related Pages:

			https://plot.ly/javascript/plotlyjs-function-reference/#common-parameters
    https://plot.ly/javascript/

    :param record:
    :param profile:
    :param width:
    :param height:
    :param htmlCode:

    :return:

    :rtype: graph.GraphPlotly.Chart
    """
    data = []
    histo_axis = ('y', y_columns) if y_columns is not None else ('x', x_columns)
    for rec in record:
      for y in histo_axis[1]:
        series = {histo_axis[0]: []}
        if y in rec:
          series[histo_axis[0]].append(float(rec[y]))
        data.append(series)
    histo_chart = graph.GraphPlotly.Bar(self.parent.context.rptObj, width, height, options or {}, htmlCode, profile)
    for d in data:
      histo_chart.add_trace(d, type='histogram')
    return histo_chart

  def pie(self, record=None, y_columns=None, x_axis=None,  profile=None, options=None, width=(100, "%"), height=(330, "px"), htmlCode=None):
    """
    Description:
    ------------

    Related Pages:

			https://plotly.com/javascript/pie-charts/

    Attributes:
    ----------
    :param record: List of dict. The Python recordset
    :param y_columns: List. The columns corresponding to keys in the dictionaries in the record
    :param x_axis: String. The column corresponding to a key in the dictionaries in the record
    :param profile:
    :param options:
    :param width: Tuple. The width of the component in the page, default (100, '%')
    :param height: Tuple. The height of the component in the page, default (330, "px")
    :param htmlCode:
    """
    options = options or {}
    options.update({'y_columns': y_columns, 'x_column': x_axis, 'type': 'pie', 'marker': {'colors': self.parent.context.rptObj.theme.charts}, 'mode': None, 'attrs': {'orientation': 'h'}})
    data = self.parent.context.rptObj.data.plotly.xy(record, y_columns, x_axis)
    pie_chart = graph.GraphPlotly.Pie(self.parent.context.rptObj, width, height, options or {}, htmlCode, profile)
    for d in data['datasets']:
      pie_chart.add_trace({"label": d['x'], "values": d['y']})
      pie_chart.data.marker.colors = self.parent.context.rptObj.theme.charts
    return pie_chart

  def area(self, record, y_columns=None, x_axis=None, profile=None, options=None, width=(100, "%"), height=(330, "px"), htmlCode=None):
    """
    Description:
    ------------
    How to make a D3.js-based filled area plot in javascript. An area chart displays a solid color between the traces of a graph.

    Related Pages:

			https://plotly.com/javascript/filled-area-plots/

    Attributes:
    ----------
    :param record: List of dict. The Python recordset
    :param y_columns: List. The columns corresponding to keys in the dictionaries in the record
    :param x_axis: String. The column corresponding to a key in the dictionaries in the record
    :param profile:
    :param options:
    :param width: Tuple. The width of the component in the page, default (100, '%')
    :param height: Tuple. The height of the component in the page, default (330, "px")
    :param htmlCode:
    """
    options = options or {}
    options.update({'y_columns': y_columns, 'x_column': x_axis, 'type': "scatter", 'attrs': {'fill': "tozeroy"}})
    data = self.parent.context.rptObj.data.plotly.xy(record, y_columns, x_axis)
    line_chart = graph.GraphPlotly.Line(self.parent.context.rptObj, width, height, options, htmlCode, profile)
    for d in data['datasets']:
      line_chart.add_trace(d)
      line_chart.data.type = options['type']
      line_chart.data.fill = "tozeroy"
    return line_chart

  def bubble(self, record, y_columns=None, x_axis=None, profile=None, options=None, width=(100, "%"), height=(330, "px"), htmlCode=None):
    """
    Description:
    ------------
    How to make a D3.js-based filled area plot in javascript. An area chart displays a solid color between the traces of a graph.

    Related Pages:

			https://plotly.com/javascript/bubble-charts/

    Attributes:
    ----------
    :param record: List of dict. The Python recordset
    :param y_columns: List. The columns corresponding to keys in the dictionaries in the record
    :param x_axis: String. The column corresponding to a key in the dictionaries in the record
    :param profile:
    :param options:
    :param width: Tuple. The width of the component in the page, default (100, '%')
    :param height: Tuple. The height of the component in the page, default (330, "px")
    :param htmlCode:
    """
    options = options or {}
    options.update({'y_columns': y_columns, 'x_column': x_axis, 'mode': 'markers'})
    data = self.parent.context.rptObj.data.plotly.xy(record, y_columns, x_axis)
    line_chart = graph.GraphPlotly.Line(self.parent.context.rptObj, width, height, options, htmlCode, profile)
    for d in data['datasets']:
      line_chart.add_trace(d, mode=options['mode'])
    return line_chart

  def number(self, value, profile=None, options=None, width=(100, "%"), height=(330, "px"), htmlCode=None):
    """
    Description:
    ------------

    Related Pages:

			https://plotly.com/javascript/indicator/

    Attributes:
    ----------
    :param value:
    :param profile:
    :param options:
    :param width: Tuple. The width of the component in the page, default (100, '%')
    :param height: Tuple. The height of the component in the page, default (330, "px")
    :param htmlCode:
    """
    options = options or {}
    options.update({'type': 'indicator', 'mode': "number"})
    ind = graph.GraphPlotly.Indicator(self.parent.context.rptObj, width, height, options, htmlCode, profile)
    ind.add_trace({'value': value}, mode=options["mode"])
    return ind

  def number_with_delta(self, value, delta=100, profile=None, options=None, width=(100, "%"), height=(330, "px"), htmlCode=None):
    """
    Description:
    ------------

    Related Pages:

			https://plotly.com/javascript/indicator/

    Attributes:
    ----------
    :param value:
    :param profile:
    :param options:
    :param width: Tuple. The width of the component in the page, default (100, '%')
    :param height: Tuple. The height of the component in the page, default (330, "px")
    :param htmlCode:
    """
    options = options or {}
    options.update({'type': 'indicator', 'mode': "number+delta", "delta": {'reference': delta}})
    ind = graph.GraphPlotly.Indicator(self.parent.context.rptObj, width, height, options, htmlCode, profile)
    ind.add_trace({'value': value}, mode=options["mode"])
    ind.data.delta.reference = delta
    return ind

  def gauge(self, value, profile=None, options=None, width=(100, "%"), height=(330, "px"), htmlCode=None):
    """
    Description:
    ------------
    How to make a D3.js-based gauge chart in javascript.


    gauge = rptObj.ui.charts.plotly.gauge(2000)
    gauge.data.gauge.axis.range = [0, 5000]

    Related Pages:

			https://plotly.com/javascript/gauge-charts/

    Attributes:
    ----------
    :param value:
    :param profile:
    :param options:
    :param width: Tuple. The width of the component in the page, default (100, '%')
    :param height: Tuple. The height of the component in the page, default (330, "px")
    :param htmlCode:
    """
    options = options or {}
    options.update({'type': 'indicator', 'mode': "gauge+number"})
    gau = graph.GraphPlotly.Indicator(self.parent.context.rptObj, width, height, options or {}, htmlCode, profile)
    gau.add_trace({'value': value}, mode=options['mode'], type=options['type'])
    return gau

  def scatterpolar(self, records, r_columns=None, theta_axis=None, profile=None, options=None, width=(100, "%"), height=(330, "px"), htmlCode=None):
    all_series = []
    for c in r_columns:
      series = {"r": [], "theta": []}
      for rec in records:
        series['r'].append(rec[c])
        series['theta'].append(rec[theta_axis])
      all_series.append(series)
    spolar_chart = graph.GraphPlotly.ScatterPolar(self.parent.context.rptObj, width, height, options or {}, htmlCode, profile)
    for d in all_series:
      spolar_chart.add_trace(d, mode="line")
      spolar_chart.data.marker.color = None
    return spolar_chart

  def box(self, records, y_columns=None, x_columns=None, profile=None, options=None, width=(100, "%"), height=(330, "px"), htmlCode=None):

    axis, cols = ('y', y_columns) if y_columns is not None else ('x', x_columns)
    series = []
    for c in cols:
      series.append([rec.get(c) for rec in records])
    box_chart = graph.GraphPlotly.Box(self.parent.context.rptObj, width, height, options or {}, htmlCode, profile)
    for s in series:
      box_chart.add_trace({axis: s})
    return box_chart

  def group_box(self, records, y_columns=None, x_axis=None, profile=None, options=None, width=(100, "%"), height=(330, "px"), htmlCode=None):
    """
    https://plot.ly/javascript/box-plots/

    :param records:
    :param y_columns:
    :param x_axis:
    :param profile:
    :param options:
    :param width:
    :param height:
    :param htmlCode:
    """
    series, x = [[] for _ in range(len(y_columns))], []
    for rec in records:
      for i, c in enumerate(y_columns):
        series[i].append(rec.get(c))
      x.append(rec.get(x_axis))
    box_chart = graph.GraphPlotly.Box(self.parent.context.rptObj, width, height, options or {}, htmlCode, profile)
    for s in series:
      box_chart.add_trace({'y': s, 'x': x})
    return box_chart

  def candlestick(self, records, closes, highs, lows, opens, x_axis, profile=None, options=None, width=(100, "%"), height=(330, "px"), htmlCode=None):
    """

    data = rptObj.py.requests.csv(data_urls.PLOTLY_APPLE_PRICES)
    sc = rptObj.ui.charts.plotly.candlestick(data, closes=["AAPL.Close"], highs=["AAPL.High"], lows=["AAPL.Low"], opens=["AAPL.Open"], x_axis='Date')

    :param records:
    :param closes:
    :param highs:
    :param lows:
    :param opens:
    :param x_axis:
    :param profile:
    :param options:
    :param width:
    :param height:
    :param htmlCode:
    :return:
    """
    all_series = []
    for i, c in enumerate(closes):
      series = {'x': [], 'close': [], 'high': [], 'low': [], 'open': []}
      for rec in records:
        series['x'].append(rec[x_axis])
        series['close'].append(rec[c])
        series['high'].append(rec[highs[i]])
        series['low'].append(rec[lows[i]])
        series['open'].append(rec[opens[i]])
      all_series.append(series)

    candle_chart = graph.GraphPlotly.CandleStick(self.parent.context.rptObj, width, height, options or {}, htmlCode, profile)
    for s in all_series:
      candle_chart.add_trace(s)
    candle_chart.layout.no_background()
    return candle_chart


class Plotly3D(object):
  def __init__(self, context):
    self.parent = context
    self.chartFamily = "Plotly"

  def scatter(self, record, y_columns=None, x_axis=None, z_axis=None, profile=None, options=None, width=(100, "%"),
              height=(500, "px"), htmlCode=None):
    """

    https://plot.ly/javascript/3d-line-plots/

    :param record:
    :param y_columns:
    :param x_axis:
    :param z_axis:
    :param profile:
    :param options:
    :param width:
    :param height:
    :param htmlCode:
    """
    options = options or {}
    options.update({'y_columns': y_columns, 'x_column': x_axis, 'z_axis': z_axis, 'type': 'scatter3d', 'mode': 'markers'})
    data = self.parent.context.rptObj.data.plotly.xyz(record, y_columns, x_axis, z_axis)
    sc_chart = graph.GraphPlotly.Scatter3D(self.parent.context.rptObj, width, height, options or {}, htmlCode, profile)
    for i, series in enumerate(data['datasets']):
      sc_chart.add_trace({'x': series['x'], 'y': series['y'], 'z': series['z']})
      sc_chart.data.line.color = self.parent.context.rptObj.theme.colors[i]
    return sc_chart

  def line(self, record, y_columns=None, x_axis=None, z_axis=None, profile=None, options=None, width=(100, "%"),
              height=(500, "px"), htmlCode=None):
    """

    https://plot.ly/javascript/3d-line-plots/

    :param record:
    :param y_columns:
    :param x_axis:
    :param z_axis:
    :param profile:
    :param options:
    :param width:
    :param height:
    :param htmlCode:
    """
    options = options or {}
    options.update({'y_columns': y_columns, 'x_column': x_axis, 'z_axis': z_axis, 'type': 'scatter3d', 'mode': 'lines'})
    data = self.parent.context.rptObj.data.plotly.xyz(record, y_columns, x_axis, z_axis)
    sc_chart = graph.GraphPlotly.Scatter3D(self.parent.context.rptObj, width, height, options or {}, htmlCode, profile)
    for i, series in enumerate(data['datasets']):
      sc_chart.add_trace({'x': series['x'], 'y': series['y'], 'z': series['z']})
      #sc_chart.data.line.color = self.parent.context.rptObj.theme.colors[i]
    return sc_chart

  def marker(self, record, y_columns=None, x_axis=None, z_axis=None, profile=None, options=None, width=(100, "%"),
              height=(500, "px"), htmlCode=None):
    """

    https://plot.ly/javascript/3d-line-plots/

    :param record:
    :param y_columns:
    :param x_axis:
    :param z_axis:
    :param profile:
    :param options:
    :param width:
    :param height:
    :param htmlCode:
    """
    options = options or {}
    options.update({'y_columns': y_columns, 'x_column': x_axis, 'z_axis': z_axis, 'type': 'scatter3d', 'mode': 'lines+markers'})
    data = self.parent.context.rptObj.data.plotly.xyz(record, y_columns, x_axis, z_axis)
    sc_chart = graph.GraphPlotly.Scatter3D(self.parent.context.rptObj, width, height, options or {}, htmlCode, profile)
    for i, series in enumerate(data['datasets']):
      sc_chart.add_trace({'x': series['x'], 'y': series['y'], 'z': series['z']})
      sc_chart.data.line.color = self.parent.context.rptObj.theme.colors[i]
    return sc_chart

  def ribbon(self, record, y_columns=None, x_axis=None, z_axis=None, profile=None, options=None, width=(100, "%"),
             height=(500, "px"), htmlCode=None):
    """
    Description:
    ------------
    Create ribbons on the x axis

    Related Pages:

      https://plot.ly/javascript/ribbon-plots/

    ttributes:
    ----------
    :param record:
    :param y_columns:
    :param x_axis:
    :param z_axis:
    :param profile:
    :param options:
    :param width:
    :param height:
    :param htmlCode:
    """
    options = options or {'delta': {"x": 1, 'y': 1, 'z': 0}}
    options.update({'y_columns': y_columns, 'x_column': x_axis, 'z_axis': z_axis, 'type': 'scatter3d', 'mode': 'lines+markers'})
    data = self.parent.context.rptObj.data.plotly.x_yz(record, y_columns, x_axis, z_axis, dy=options['delta']['y'], dx=options['delta']['x'], dz=options['delta']['z'])
    line_chart = graph.GraphPlotly.Surface(self.parent.context.rptObj, width, height, options or {}, htmlCode,
                                           profile)
    for i, d in enumerate(data['datasets']):
      line_chart.add_trace(d)
      line_chart.data.showscale = False
    return line_chart

  def mesh3d(self, records, intensity, x, y, z, i=None, j=None, k=None, profile=None, options=None, width=(100, "%"),
             height=(500, "px"), htmlCode=None):
    """

    https://plot.ly/javascript/3d-mesh/

    :param records:
    :param intensity:
    :param x:
    :param y:
    :param z:
    :param i:
    :param j:
    :param k:
    :param profile:
    :param options:
    :param width:
    :param height:
    :param htmlCode:
    """
    data = {"intensity": [], 'x': [], 'y': [], 'z': []}
    if i is not None:
      data[i] = []
    if j is not None:
      data[j] = []
    if k is not None:
      data[k] = []
    for rec in records:
      data["intensity"].append(rec[intensity])
      data["x"].append(rec[x])
      data["y"].append(rec[y])
      data["z"].append(rec[z])
      if i is not None:
        data["i"].append(rec[i])
      if j is not None:
        data["j"].append(rec[j])
      if k is not None:
        data["k"].append(rec[k])
    mesh_chart = graph.GraphPlotly.Mesh3d(self.parent.context.rptObj, width, height, options or {}, htmlCode, profile)
    mesh_chart.add_trace(data)
    return mesh_chart

  def surface(self, record, y_columns=None, x_axis=None, z_axis=None, profile=None, options=None, width=(100, "%"),
              height=(500, "px"), htmlCode=None):
    """

    :param record:
    :param y_columns:
    :param x_axis:
    :param z_axis:
    :param profile:
    :param options:
    :param width:
    :param height:
    :param htmlCode:
    """
    options = options or {}
    options.update({'type': 'surface', 'mode': ''})
    naps = self.parent.context.rptObj.data.plotly.surface(record, y_columns, x_axis, z_axis)
    surf_chart = graph.GraphPlotly.Surface(self.parent.context.rptObj, width, height, options, htmlCode, profile)
    for i, d in enumerate(naps['datasets']):
      surf_chart.add_trace({'z': d})
      surf_chart.data.showscale = False
    return surf_chart

  def maps(self, records, profile=None, options=None, width=(100, "%"), height=(500, "px"), htmlCode=None):

    options = options or {}
    options.update({'type': 'surface', 'mode': ''})
    surf_chart = graph.GraphPlotly.Surface(self.parent.context.rptObj, width, height, options, htmlCode, profile)
    for d in records:
      surf_chart.add_trace({'z': d})
    return surf_chart


class Plotly(Plotly2D):

  def __init__(self, context):
    super(Plotly, self).__init__(context)
    self._3d = Plotly3D(context)

  def scatter3d(self, record, y_columns=None, x_axis=None, z_axis=None, profile=None, options=None, width=(100, "%"), height=(400, "px"), htmlCode=None):
    return self._3d.scatter(record, y_columns=y_columns, x_axis=x_axis, z_axis=z_axis, profile=profile, options=options, width=width, height=height, htmlCode=htmlCode)

  def line3d(self, record, y_columns=None, x_axis=None, z_axis=None, profile=None, options=None, width=(100, "%"), height=(400, "px"), htmlCode=None):
    return self._3d.line(record, y_columns=y_columns, x_axis=x_axis, z_axis=z_axis, profile=profile, options=options, width=width, height=height, htmlCode=htmlCode)

  def marker3d(self, record, y_columns=None, x_axis=None, z_axis=None, profile=None, options=None, width=(100, "%"), height=(400, "px"), htmlCode=None):
    return self._3d.marker(record, y_columns=y_columns, x_axis=x_axis, z_axis=z_axis, profile=profile, options=options, width=width, height=height, htmlCode=htmlCode)

  def ribbon(self, record, y_columns=None, x_axis=None, z_axis=None, profile=None, options=None, width=(100, "%"), height=(500, "px"), htmlCode=None):
    return self._3d.ribbon(record, y_columns, x_axis, z_axis, profile, options, width, height, htmlCode)

  def mesh3d(self, record, intensity, x, y, z, i=None, j=None, k=None, profile=None, options=None, width=(100, "%"), height=(500, "px"), htmlCode=None):
    return self._3d.mesh3d(record, intensity, x, y, z, i, j, k, profile, options, width, height, htmlCode)

  def surface(self, record, y_columns=None, x_axis=None, z_axis=None, profile=None, options=None, width=(100, "%"), height=(500, "px"), htmlCode=None):
    return self._3d.surface(record, y_columns, x_axis, z_axis, profile, options, width, height, htmlCode)

  def maps(self, record, profile=None, options=None, width=(100, "%"), height=(500, "px"), htmlCode=None):
    return self._3d.maps(record, profile, options, width, height, htmlCode)



"""
Module dedicated to send mails
"""

import email
import email.mime
import email.mime.image
import email.mime.text
import email.mime.application
import email.mime.multipart
import email.mime.message
import smtplib
import sys

_DEBUG = False
_PY3 = sys.version_info.major == 3
_STRING_TYPES = (str,) if _PY3 else (str, unicode)
_str = str if _PY3 else unicode


class Attachment(object):
  """Email attachment class.

  This instantiates a proxy object that will only read
  the data whenever the `data` property is read. This
  class is not suitable for huge files.

  INSTANCE PROPERTIES
  -------------------
  filename (rw): The filename to give to the attachment
  fileobj (rw): The file object to read the data from
  data (r): The data read from the file object
  mimetype (rw): The MIME type of the data
  """
  def __init__(self, filename, fileobj, mimetype='application/octet-stream'):
    """Create an attachment object

    ARGUMENTS
    ---------
    (See instance properties)
    """
    self.filename = filename
    self.fileobj = fileobj
    self.mimetype = mimetype
    self.__data = None # Raw file data

  @property
  def data(self):
    if self.__data is None:
      self.__data = self.fileobj.read()

    return self.__data


class Address(object):
  """Represents an email address.

  This class can be used to add email addresses to emails with attached name
  information.
  """
  def __init__(self, email, name=None):
    self.name = name
    self.email = email

  def __str__(self):
    return _str(
      '%s%s' % (
        (('%s ' % self.name) if self.name is not None else ''),
        '<%s>' % self.email
      )
    )

  def __eq__(self, other):
    if isinstance(other, _STRING_TYPES):
      return self.email == other or _str(Address(self.email)) == other or _str(self) == other
    elif isinstance(other, self.__class__):
      return self.email == other.email and self.name == other.name
    else:
      raise ValueError('Only string and Address objects can be compared with an Address object')


class Email(object):
  def __init__(self, sender, recipients, subject, content, attachments=None, headers=None, mimetype='text/plain', charset='utf8'):
    """
    ARGUMENTS
    ---------
    sender: Email sender
    recipients: Iterable object of recipients (can be both strings or Address objects)
    subject: Email subject
    content: Email content
    attachments: Optional iterable of Attachment objects
    headers: Optional dictionary of additional headers in the form {header_name: header_value}
    mimetype: The content type of the message
    charset: Optional charset
    """
    self.sender = sender
    self.recipients = tuple(map(lambda r: r if isinstance(r, Address) else Address(r), recipients))
    self.subject = subject
    self.content = content
    self.attachments = tuple(attachments or ()) # TODO: Test attachment file names
    self.headers = tuple(headers or ())
    self.mimetype = mimetype
    self.charset = charset

  def to_mime(self):
    """Return a MIME representation of this object - used internally"""
    e = email.mime.multipart.MIMEMultipart()

    e.set_charset(self.charset)
    e['Subject'] = self.subject

    maintype, subtype = self.mimetype.lower().split('/', 1)
    m = email.mime.text.MIMEText(self.content, _subtype=subtype, _charset=self.charset)
    e.attach(m)

    for h, hv in self.headers:
      e[h] = hv
    for a in self.attachments:
      maintype, subtype = a.mimetype.lower().split('/', 1)
      if maintype == 'image':
        aclass = email.mime.image.MIMEImage
      else:
        aclass = email.mime.application.MIMEApplication
      mimea = aclass(a.data, _subtype=subtype)
      mimea.add_header('Content-Disposition', 'attachment', filename=a.filename)

      e.attach(mimea)

    return e


class SMTPServer(object):
  """SMTP server class.

  This class is used to represent a SMTP server and a connection to it.
  """
  ENCRYPTION_METHODS = ('tls', 'ssl', '') # Supported connection encryption methods

  def __init__(self, hostname, port, encryption='tls'):
    """Initialise a new instance of this class.

    ARGUMENTS
    ---------
    hostname: SMTP server hostname
    port: SMTP server port
    encryption: String specifying the form of encryption required for the connection
      Must be one of `Server.ENCRYPTION_METHODS`
    """
    self.hostname = hostname
    self.port = port
    self.encryption = encryption.lower()
    self.server = None

    if self.encryption not in self.__class__.ENCRYPTION_METHODS:
      raise ValueError('encryption must be one of %s', repr(self.__class__.ENCRYPTION_METHODS))

  def connect(self, **kwargs):
    """Connect to a SMTP server.

    OPTIONAL KEYWORD ARGUMENTS
    -----------------
    keyfile, certfile, context: Key/certificate file and TLS context for the STARTTLS method
    user, password: Login credentials
    """
    ssl = self.encryption == 'ssl'
    tls = self.encryption == 'tls'
    cls = smtplib.SMTP if not ssl else smtplib.SMTP_SSL
    srv = cls(host=self.hostname, port=self.port)

    if _DEBUG:
      srv.set_debuglevel(1)

    if tls:
      srv.starttls(
        kwargs.get('keyfile', None),
        kwargs.get('certfile', None),
        kwargs.get('context', None)
      )

    try:
      user = kwargs['user']
      password = kwargs['password']

      srv.login(user, password)
    except KeyError:
      pass # No login information provided

    self.server = srv

  def disconnect(self):
    """Disconnect from the SMTP server"""
    if self.is_connected:
      self.server.quit()
      self.server = None

  @property
  def is_connected(self):
    """Evaluates to `True` if the object represents a connected instance"""
    return self.server is not None

  def sendmail(self, email, **kwargs):
    """Send an email.

    ARGUMENTS
    ---------
    email: The `Email` object to send

    OPTIONAL KEYWORD ARGUMENTS
    -----------------
    (Directly passed to `server.sendmessage()`)
    """
    if not self.is_connected:
      self.connect()

    self.server.sendmail(
      _str(email.sender),
      tuple(_str(x) for x in email.recipients),
      email.to_mime().as_string(),
      **kwargs
    )




from epyk.core.html import geo


class Dc(object):
  def __init__(self, context):
    self.parent = context
    self.chartFamily = "DC"

  def usa(self, record=None, y_columns=None, x_axis=None, title=None, profile=None, options=None, width=(100, "%"),
           height=(330, "px"), htmlCode=None):
    """
    Description:
    -----------

    Underlying HTML Objects:

      - :class:`epyk.core.geo.GeoDc.ChartGeoChoroplethk`

    Related Pages:

			https://jsfiddle.net/djmartin_umich/9VJHe/
      http://bl.ocks.org/KatiRG/cccd23dd7a830da0de5c

    Attributes:
    ----------
    :param record:
    :param y_columns:
    :param x_axis:
    :param title:
    :param profile:
    :param width:
    :param height:
    :param htmlCode:
    """
    geo_chart = geo.GeoDc.ChartGeoChoropleth(self.parent.context.rptObj, width, height, title, options or {}, htmlCode, profile)
    line_id = geo_chart.htmlCode
    self.parent.context.rptObj._props.setdefault('js', {}).setdefault('datasets', {})['data_cf_%s' % line_id] = "var %(cId)s_cf = crossfilter(%(data)s); var %(cId)s_dim = %(cId)s_cf.dimension(function(d) {return +d['%(x)s'];})" % {'cId': line_id, 'data': record, 'x': x_axis}
    geo_chart.dom.height(height[0]).dimension("%s_dim" % line_id).group("%(cId)s_dim.group().reduceSum(function(d) {return d['%(y)s'] ;})" % {'cId': line_id, 'y': y_columns})
    return geo_chart



""" .. _CompButtons:

Button Interfaces
=================

"""
# Check if pandas is available in the current environment
# if it is the case this module can handle DataFrame directly
try:
  import pandas as pd
  has_pandas = True

except:
  has_pandas = False

from epyk.core import html


class Buttons(object):
  def __init__(self, context):
    self.context = context

  def _recordSet(self, recordSet, column):
    """
    Attributes:
    ----------
    :param recordSet:
    :param column:
    :return:
    """
    data = None
    is_converted = False
    if has_pandas:
      if isinstance(recordSet, pd.DataFrame):
        data = [{'name': r, 'value': r} for r in recordSet[column].unique().tolist()]
        is_converted = True

    if not is_converted:
      result = {}
      for rec in recordSet:
        result[rec[column]] = {'name': rec[column], 'value': rec[column]}
      data = [result[k] for k in sorted(result.keys())]
    return data

  def button(self, text="", icon=None, width=(None, "%"), height=(None, "px"), htmlCode=None, tooltip=None, profile=None, options=None):
    """
    Description:
    ------------
    Standard button

    Usage::

      rptObj.ui.button("Test")


    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlButton.Button`

    Related Pages:

      https://www.w3schools.com/tags/tag_button.asp
      http://www.kodingmadesimple.com/2015/04/custom-twitter-bootstrap-buttons-icons-images.html

    Attributes:
    ----------
    :param text: Optional. The value to be displayed to the button
    :param width: Optional. A tuple with the integer for the component width and its unit
    :param height: Optional. A tuple with the integer for the component height and its unit
    :param icon: Optional. A string with the value of the icon to display from font-awesome
    :param htmlCode: Optional. An identifier for this component (on both Python and Javascript side)
    :param tooltip: Optional. A string with the value of the tooltip
    :param profile: Optional. A flag to set the component performance storage
    :param options: Optional. Specific Python options available for this component

    :return: The Button HTML object
    """
    html_button = html.HtmlButton.Button(self.context.rptObj, text, icon, width, height, htmlCode=htmlCode,
                                         tooltip=tooltip, profile=profile, options=options)
    return html_button

  def small(self, text="", icon=None, width=(None, "%"), height=(None, "px"), htmlCode=None, tooltip=None,
             profile=None, options=None):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param text:
    :param icon:
    :param width:
    :param height:
    :param htmlCode:
    :param tooltip:
    :param profile:
    :param options:
    """
    html_button = html.HtmlButton.Button(self.context.rptObj, text, icon, width, height, htmlCode=htmlCode,
                                         tooltip=tooltip, profile=profile, options=options)
    html_button.style.css.line_height = 12
    html_button.style.css.padding = 2
    return html_button

  def important(self, text="", icon=None, width=(None, "%"), height=(None, "px"), htmlCode=None, tooltip=None, profile=None, options=None):
    """
    Description:
    ------------
    Same as Standard button but used to attract user attention

    Usage::

      rptObj.ui.buttons.important("Important")

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlButton.Button`

    Related Pages:

      https://www.w3schools.com/tags/tag_button.asp
      http://www.kodingmadesimple.com/2015/04/custom-twitter-bootstrap-buttons-icons-images.html

    Attributes:
    ----------
    :param text: Optional. The value to be displayed to the button
    :param width: Optional. A tuple with the integer for the component width and its unit
    :param height: Optional. A tuple with the integer for the component height and its unit
    :param icon: Optional. A string with the value of the icon to display from font-awesome
    :param htmlCode: Optional. An identifier for this component (on both Python and Javascript side)
    :param tooltip: Optional. A string with the value of the tooltip
    :param profile: Optional. A flag to set the component performance storage
    :param options: Optional. Specific Python options available for this component

    :return: The Button HTML object
    """
    html_button = html.HtmlButton.Button(self.context.rptObj, text, icon, width, height, htmlCode=htmlCode,
                                         tooltip=tooltip, profile=profile, options=options)
    html_button.style.add_classes.button.important()
    return html_button

  def validate(self, text="", width=(None, "%"), height=(None, "px"), htmlCode=None, tooltip=None, profile=None, options=None):
    """
    Description:
    -----------

    Usage::

      rptObj.ui.buttons.validate()

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlButton.Button`

    Related Pages:

        https://www.w3schools.com/tags/tag_button.asp
        http://www.kodingmadesimple.com/2015/04/custom-twitter-bootstrap-buttons-icons-images.html

    Attributes:
    ----------
    :param text: Optional. The value to be displayed to the button
    :param width: Optional. A tuple with the integer for the component width and its unit
    :param height: Optional. A tuple with the integer for the component height and its unit
    :param htmlCode: Optional. An identifier for this component (on both Python and Javascript side)
    :param tooltip: Optional. A string with the value of the tooltip
    :param profile: Optional. A flag to set the component performance storage
    :param options: Optional. Specific Python options available for this component
    """
    html_but = html.HtmlButton.Button(self.context.rptObj, text, 'fas fa-check-circle', width, height, htmlCode=htmlCode,
                                      tooltip=tooltip, profile=profile, options=options)
    return html_but

  def remove(self, text="", width=(None, "%"), height=(None, "px"), htmlCode=None, tooltip=None, profile=None, options=None):
    """
    Description:
    -----------
    Button with cross icon

    Usage::

      rptObj.ui.buttons.remove()

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlButton.Button`

    Related Pages:

			https://www.w3schools.com/tags/tag_button.asp
    http://www.kodingmadesimple.com/2015/04/custom-twitter-bootstrap-buttons-icons-images.html

    Attributes:
    ----------
    :param text: Optional. The value to be displayed to the button
    :param width: Optional. A tuple with the integer for the component width and its unit
    :param height: Optional. A tuple with the integer for the component height and its unit
    :param htmlCode: Optional. An identifier for this component (on both Python and Javascript side)
    :param tooltip: Optional. A string with the value of the tooltip
    :param profile: Optional. A flag to set the component performance storage
    :param options: Optional. Specific Python options available for this component

    :return: The Button HTML object
    """
    html_but = html.HtmlButton.Button(self.context.rptObj, text, 'fas fa-trash-alt', width, height, htmlCode=htmlCode,
                             tooltip=tooltip, profile=profile, options=options)
    return html_but

  def phone(self, text="", width=(None, "%"), height=(None, "px"), htmlCode=None, tooltip=None, profile=None, options=None):
    """
    Description:
    -----------

    Usage::

      rptObj.ui.buttons.phone()

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlButton.Button`

    Related Pages:

			https://www.w3schools.com/tags/tag_button.asp
    http://www.kodingmadesimple.com/2015/04/custom-twitter-bootstrap-buttons-icons-images.html

    Attributes:
    ----------
    :param text: Optional. The value to be displayed to the button
    :param width: Optional. A tuple with the integer for the component width and its unit
    :param height: Optional. A tuple with the integer for the component height and its unit
    :param htmlCode: Optional. An identifier for this component (on both Python and Javascript side)
    :param tooltip: Optional. A string with the value of the tooltip
    :param profile: Optional. A flag to set the component performance storage
    :param options: Optional. Specific Python options available for this component

    :return: The Button HTML object
    """
    html_button = html.HtmlButton.Button(self.context.rptObj, text, 'fas fa-phone', width, height, htmlCode=htmlCode,
                                         tooltip=tooltip, profile=profile, options=options)
    return html_button

  def mail(self, text="", width=(None, "%"), height=(None, "px"), htmlCode=None, tooltip=None, profile=None, options=None):
    """
    Description:
    ------------

    Usage::

      rptObj.ui.buttons.mail()

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlButton.Button`

    Related Pages:

			https://www.w3schools.com/tags/tag_button.asp
    http://www.kodingmadesimple.com/2015/04/custom-twitter-bootstrap-buttons-icons-images.html

    Attributes:
    ----------
    :param text: Optional. The value to be displayed to the button
    :param width: Optional. A tuple with the integer for the component width and its unit
    :param height: Optional. A tuple with the integer for the component height and its unit
    :param htmlCode: Optional. An identifier for this component (on both Python and Javascript side)
    :param tooltip: Optional. A string with the value of the tooltip
    :param profile: Optional. A flag to set the component performance storage
    :param options: Optional. Specific Python options available for this component
    """
    html_but = html.HtmlButton.Button(self.context.rptObj, text, 'fas fa-envelope', width, height, htmlCode=htmlCode,
                                      tooltip=tooltip, profile=profile, options=options)
    return html_but

  def radio(self, recordSet=None, checked=None, htmlCode=None, label=None, width=(100, '%'), height=(None, "px"), radioVisible=False,
            event=None, withRemoveButton=False, column=None, align='left', filters=None, tooltip='', allSelected=False,
            radioType="row", helper=None, profile=None):
    """
    Description:
    ------------
    Creates a radio HTML component

    Usage::

      rptObj.ui.buttons.radio(df, dfColumn="A", htmlCode="test")

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlRadio.Radio`

    Related Pages:

			https://www.w3schools.com/bootstrap/bootstrap_forms_inputs.asp

    Attributes:
    ----------
    :param recordSet:
    :param checked:
    :param htmlCode:
    :param label:
    :param width: Optional. Integer for the component width
    :param height: Optional. Integer for the component height
    :param radioVisible:
    :param event:
    :param withRemoveButton:
    :param column:
    :param align:
    :param filters:
    :param tooltip:
    :param allSelected:
    :param radioType:
    :param profile:
    """
    recordSet = recordSet or []

    if column is not None:
      recordSet = self._recordSet(recordSet, column)
    if isinstance(recordSet, list) and recordSet and not isinstance(recordSet[0], dict):
      tmpVals = [{'value': str(v)} for v in recordSet]
      tmpVals[0]['checked'] = True
      recordSet = tmpVals
    html_radio = html.HtmlRadio.Radio(self.context.rptObj, recordSet, htmlCode, label, width,
                      height, radioVisible, event, withRemoveButton, align, filters, tooltip, radioType, helper, profile)
    return html_radio

  def toggle(self, recordSet=None, label=None, color=None, width=(None, '%'), height=(20, 'px'), htmlCode=None, profile=None):
    """
    Description:
    ------------

    Usage::

      rptObj.ui.buttons.toggle({'on': "true", 'off': 'false'})

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlRadio.Switch`

    Related Pages:
http://thecodeplayer.com/walkthrough/pure-css-on-off-toggle-switch
    https://codepen.io/mburnette/pen/LxNxNg

    Attributes:
    ----------
    :param recordSet:
    :param label:
    :param color:
    :param width: Optional. Integer for the component width
    :param height: Optional. Integer for the component height
    :param htmlCode:
    :param profile:
    """
    html_toggle = html.HtmlRadio.Switch(self.context.rptObj, recordSet, label, color, width, height, htmlCode, profile)
    return html_toggle

  def checkboxes(self, records=None, title=None, color=None, width=(100, "%"), height=(None, "px"), align='left',
               htmlCode=None, tooltip='', dfColumn=None, icon="fas fa-check", options=None, profile=None):
    """
    Description:
    ------------
    Python wrapper to the HTML checkbox elements

    Usage::

      rptObj.ui.buttons.checkboxes(data)

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlButton.Checkbox`

    Related Pages:

			https://www.w3schools.com/howto/howto_css_custom_checkbox.asp

    Attributes:
    ----------
    :param records:
    :param title:
    :param color:
    :param width:
    :param height:
    :param align:
    :param htmlCode:
    :param tooltip:
    :param dfColumn:
    :param icon:
    :param profile:
    """
    if dfColumn is not None:
      if has_pandas and issubclass(type(records), pd.DataFrame):
        # if globalFilter:
        #   dataId = id(records)
        #   dataCode = "df_code_%s" % dataId
        #   globalFilter = {'jsId': dataCode, 'colName': dfColumn, 'allSelected': options.get("all_selected", False), 'operation': 'in'}
        #   if not dataCode in self.context.rptObj.jsSources:
        #     self.context.rptObj.jsSources[dataCode] = {'dataId': dataId, 'containers': [], 'data': records}
        #     self.context.rptObj.jsSources[dataCode]['containers'].append(self)
        if options.get("all_selected", False):
          records = [{"value": rec, "checked": True} for rec in records[dfColumn].unique().tolist()]
        else:
          records = records[dfColumn].unique().tolist()
    elif isinstance(records, list) and len(records) > 0:
      if not isinstance(records[0], dict):
        records = [{"value": rec} for rec in records]
    html_boxes = html.HtmlButton.Checkbox(self.context.rptObj, records, title, color, width,
                                             height, align, htmlCode, tooltip, icon, options or {}, profile)
    return html_boxes

  def check(self, flag=False, tooltip=None, width=(None, "px"), height=(20, "px"), label=None, icon=None, htmlCode=None,
            profile=None, options=None):
    """
    Description:
    ------------
    Wrapper to the check box button object

    Usage::

      rptObj.ui.buttons.check(label="Label")
      rptObj.ui.buttons.check(True, label="Label")
      rptObj.ui.buttons.check(True, label="Label", icon="fas fa-align-center")

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlButton.CheckButton`

    Related Pages:
Attributes:
    ----------
    :param flag: Optional. The value of the checkbox. Default False
    :param tooltip: Optional. A string with the value of the tooltip
    :param width: Optional. A tuple with the integer for the component width and its unit
    :param height: Optional. A tuple with the integer for the component height and its unit
    :param label: Optional. The component label content
    :param icon: Optional. The icon to be used in the check component
    :param htmlCode: Optional. An identifier for this component (on both Python and Javascript side)
    :param profile: Optional. A flag to set the component performance storage
    :param options: Optional. Specific Python options available for this component
    """
    html_but = html.HtmlButton.CheckButton(self.context.rptObj, flag, tooltip, width, height, icon, label, htmlCode, options or {}, profile)
    return html_but

  def menu(self, record, text="", icon=None, width=(None, "%"), height=(None, "px"), htmlCode=None, tooltip=None, profile=None, options=None):
    """
    Description:
    -----------
    Button with an underlying items menu

    Usage::

      tree5 = rptObj.ui.buttons.menu(["A", "B", "C"], 'Menu')

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlButton.ButtonMenu`

    Related Pages:

			https://www.w3schools.com/howto/tryit.asp?filename=tryhow_css_js_dropdown_hover

    Attributes:
    ----------
    :param record:
    :param text:
    :param icon:
    :param width:
    :param height:
    :param htmlCode:
    :param tooltip:
    :param profile:
    :param options:
    """
    html_button = html.HtmlButton.ButtonMenu(self.context.rptObj, record, text, icon, width, height, htmlCode=htmlCode,
                                              tooltip=tooltip, profile=profile, options=options)
    return html_button




from epyk.core.html import Html
from epyk.core.js import JsUtils
from epyk.core.js.packages import JsNvd3
from epyk.core.js.packages import JsD3


class Chart(Html.Html):
  name = 'NVD3 Chart'
  requirements = ('nvd3', )

  def __init__(self,  report, width, height, options, htmlCode, profile):
    self.seriesProperties, self.__chartJsEvents, self.height = {'static': {}, 'dynamic': {}}, {}, height[0]
    super(Chart, self).__init__(report, [], htmlCode=htmlCode, css_attrs={"width": width, "height": height}, profile=profile)
    self._d3, self.html_items, self._datasets = None, [], []
    self._options_init = options

  @property
  def chartId(self):
    """
    Description:
    ------------
    Return the Javascript variable of the chart
    """
    return "%s_obj" % self.htmlCode

  @property
  def data(self):
    return self._datasets[-1]

  def traces(self, i=None):
    """

    :rtype: JsChartJs.DataSetPie
    """
    if i is None:
      return self._datasets[-1]

    return self._datasets[i]

  def click(self, jsFnc, profile=False):
    raise Exception("Not implemented for this chart !")

  def add_trace(self, data, name=""):
    dataset = {"values": data, 'key': name}
    next_index = len(self._datasets)
    if len(self._report.theme.colors) > next_index:
      dataset['color'] = self._report.theme.colors[next_index]
    self._datasets.append(dataset)
    return self

  @property
  def d3(self):
    """

    :rtype: JsD3.D3Select
    """
    if self._d3 is None:
      self._d3 = JsD3.D3Select(self._report, selector="d3.select('#%s')" % self.htmlCode, setVar=False)
    return self._d3

  def convert(self, data, options, profile=False):
    mod_name = __name__.split(".")[-1]
    constructors = self._report._props.setdefault("js", {}).setdefault("constructors", {})
    constructors[self.builder_name] = "function %s%sConvert(data, options){%s; return result}" % (
      mod_name, self.builder_name, self._js__convertor__)
    if isinstance(data, dict):
      # check if there is no nested HTML components in the data
      tmp_data = ["%s: %s" % (JsUtils.jsConvertData(k, None), JsUtils.jsConvertData(v, None)) for k, v in data.items()]
      js_data = "{%s}" % ",".join(tmp_data)
    else:
      js_data = JsUtils.jsConvertData(data, None)
    options, js_options = options or self._options_init, []
    for k, v in options.items():
      if isinstance(v, dict):
        row = ["'%s': %s" % (s_k, JsUtils.jsConvertData(s_v, None)) for s_k, s_v in v.items()]
        js_options.append("'%s': {%s}" % (k, ", ".join(row)))
      else:
        if str(v).strip().startswith("function"):
          js_options.append("%s: %s" % (k, v))
        else:
          js_options.append("%s: %s" % (k, JsUtils.jsConvertData(v, None)))
    return "%s%sConvert(%s, %s)" % (mod_name, self.builder_name, js_data, "{%s}" % ",".join(js_options))

  def build(self, data=None, options=None, profile=False):
    if data:
      return "d3.select('#%(htmlCode)s').datum(%(data)s).transition().duration(500).call(%(chart)s); nv.utils.windowResize(%(chart)s.update)" % {'htmlCode': self.htmlCode, 'data': self.convert(data, options, profile), 'chart': self.dom.var}

    return JsUtils.jsConvertFncs([self.dom.set_var(True), self.dom.xAxis, self.d3.datum(self._datasets).call(self.dom.var),
                "nv.utils.windowResize(function() { %s.update() })" % self.dom.var], toStr=True)[4:]

  def __str__(self):
    self._report._props.setdefault('js', {}).setdefault("builders", []).append(self.refresh())
    str_items = "".join([h.html() for h in self.html_items])
    return '%s<svg %s></svg>' % (str_items, self.get_attrs(pyClassNames=self.style.get_classes()))


class ChartLine(Chart):

  @property
  def dom(self):
    """
    :rtype: JsNvd3.JsNvd3Line
    """
    if self._dom is None:
      self._dom = JsNvd3.JsNvd3Line(self._report, varName=self.chartId)
    return self._dom

  @property
  def _js__convertor__(self):
    return '''
      if(data.python){
        result = [];
        data.datasets.forEach(function(rec, i){
          result.push( {key: data.series[i], values: rec, labels: data.labels} )})
      } else {
        var temp = {}; var labels = []; var uniqLabels = {};
        options.y_columns.forEach(function(series){temp[series] = {}}) ;
        data.forEach(function(rec){ 
          options.y_columns.forEach(function(name){
            if(rec[name] !== undefined){
              if (!(rec[options.x_column] in uniqLabels)){labels.push(rec[options.x_column]); uniqLabels[rec[options.x_column]] = true};
              temp[name][rec[options.x_column]] = rec[name]}})
        }); result = [];
        options.y_columns.forEach(function(series){
          dataSet = {key: series, values: [], labels: labels};
          labels.forEach(function(x, i){
            var value = temp[series][x]; 
            if (isNaN(value)) { value = null};
            if (value !== undefined) {dataSet.values.push({y: value, x: i, label: x})}
          }); result.push(dataSet)})
      }'''


class ChartScatter(ChartLine):

  @property
  def dom(self):
    """
    :rtype: JsNvd3.JsNvd3Line
    """
    if self._dom is None:
      self._dom = JsNvd3.JsNvd3Scatter(self._report, varName=self.chartId)
    return self._dom

  def click(self, jsFnc, profile=False):
    self.onReady("%s.scatter.dispatch.on('elementClick', function(event){ %s })" % (self.dom.varName, JsUtils.jsConvertFncs(jsFnc, toStr=True)))
    return self


class ChartCumulativeLine(ChartLine):

  @property
  def dom(self):
    """
    :rtype: JsNvd3.JsNvd3CumulativeLine
    """
    if self._dom is None:
      self._dom = JsNvd3.JsNvd3CumulativeLine(self._report, varName=self.chartId)
    return self._dom


class ChartFocusLine(Chart):

  @property
  def dom(self):
    """
    :rtype: JsNvd3.JsNvd3CumulativeLine
    """
    if self._dom is None:
      self._dom = JsNvd3.JsNvd3LineWithFocus(self._report, varName=self.chartId)
    return self._dom


class ChartBar(Chart):

  @property
  def dom(self):
    """
    :rtype: JsNvd3.JsNvd3Bar
    """
    if self._dom is None:
      self._dom = JsNvd3.JsNvd3Bar(self._report, varName=self.chartId)
    return self._dom

  def click(self, jsFnc, profile=False):
    self.onReady("%s.selectAll('.nv-bar').on('click', function(event){ %s })" % (self.d3.varId, JsUtils.jsConvertFncs(jsFnc, toStr=True)))
    return self

  @property
  def _js__convertor__(self):
    return '''
      if(data.python){
        result = [];
        data.datasets.forEach(function(rec, i){
          result.push( {key: data.series[i], values: rec, labels: data.labels} )})
      } else {
        var temp = {}; var labels = []; var uniqLabels = {};
        options.y_columns.forEach(function(series){temp[series] = {}}) ;
        data.forEach(function(rec){ 
          options.y_columns.forEach(function(name){
            if(rec[name] !== undefined){
              if (!(rec[options.x_column] in uniqLabels)){labels.push(rec[options.x_column]); uniqLabels[rec[options.x_column]] = true};
              temp[name][rec[options.x_column]] = rec[name]}})
        }); var result = [];
        options.y_columns.forEach(function(series){
          dataSet = {key: series, values: [], labels: labels};
          labels.forEach(function(x, i){
            var value = temp[series][x]; 
            if (isNaN(value)) { value = null};
            if (value !== undefined) {dataSet.values.push({y: value, x: i, label: x})}
          }); result.push(dataSet)})
      }'''


class ChartHorizontalBar(ChartBar):

  @property
  def dom(self):
    """
    :rtype: JsNvd3.JsNvd3Bar
    """
    if self._dom is None:
      self._dom = JsNvd3.JsNvd3MultiBarHorizontal(self._report, varName=self.chartId)
    return self._dom


class ChartMultiBar(Chart):

  @property
  def dom(self):
    """
    :rtype: JsNvd3.JsNvd3Bar
    """
    if self._dom is None:
      self._dom = JsNvd3.JsNvd3MultiBar(self._report, varName=self.chartId)
    return self._dom


class ChartPie(Chart):

  @property
  def dom(self):
    """
    :rtype: JsNvd3.JsNvd3Pie
    """
    if self._dom is None:
      self._dom = JsNvd3.JsNvd3Pie(self._report, varName=self.chartId)
    return self._dom

  @property
  def _js__convertor__(self):
    return '''
      if(data.python){
        data.datasets.forEach(function(dataset, i){console.log(dataset);
          result = dataset;  
        }); console.log(result)
      } else {
        var temp = {}; var labels = {};
        data.forEach(function(rec){ 
          if(!(rec[options.x_column] in temp)){temp[rec[options.x_column]] = {}};
          options.y_columns.forEach(function(name){
            labels[name] = true; if(rec[name] !== undefined) {if (!(name in temp[rec[options.x_column]])){temp[rec[options.x_column]][name] = rec[name]} else {temp[rec[options.x_column]][name] += rec[name]}}  }) ;
        });
        var labels = Object.keys(labels); result = [];
        for(var series in temp){
          var values = {y: 0, x: series};
          labels.forEach(function(label){
            if(temp[series][label] !== undefined){values.y = temp[series][label]}});
          result.push(values)}
      }'''

  def click(self, jsFnc, profile=False):
    self.onReady("%s.pie.dispatch.on('elementClick', function(event){ %s })" % (self.dom.varName, JsUtils.jsConvertFncs(jsFnc, toStr=True)))
    return self

  def add_trace(self, data, name=""):
    """

    :param data:
    :param name:
    """
    self.dom.color(self._report.theme.colors)
    self._datasets = data
    return self


class ChartArea(ChartBar):

  @property
  def dom(self):
    """
    :rtype: JsNvd3.JsNvd3Area
    """
    if self._dom is None:
      self._dom = JsNvd3.JsNvd3Area(self._report, varName=self.chartId)
    return self._dom


class ChartHistoBar(ChartBar):

  @property
  def dom(self):
    """
    :rtype: JsNvd3.JsNvd3Area
    """
    if self._dom is None:
      self._dom = JsNvd3.JsNvd3HistoricalBar(self._report, varName=self.chartId)
    return self._dom


class ChartParallelCoord(Chart):

  @property
  def dom(self):
    """
    :rtype: JsNvd3.JsNvd3ParallelCoordinates
    """
    if self._dom is None:
      self._dom = JsNvd3.JsNvd3ParallelCoordinates(self._report, varName=self.chartId)
    return self._dom

  def set_dimension_names(self, dimensions):
    """

    :param dimensions:
    """
    self.__dimensions = dimensions
    self.dom.dimensionNames(dimensions)
    return self

  def add_trace(self, data, name=""):
    """

    :param data:
    :param name:
    """
    self._datasets = data
    return self


class ChartSunbrust(Chart):

  @property
  def dom(self):
    """
    :rtype: JsNvd3.JsNvd3Sunburst
    """
    if self._dom is None:
      self._dom = JsNvd3.JsNvd3Sunburst(self._report, varName=self.chartId)
    return self._dom

  def set_rcolors(self, color, data):
    """

    :param color:
    :param data:
    """
    for rec in data:
      rec['color'] = color
      if 'children' in rec:
        self.set_rcolors(color, rec['children'])

  def add_trace(self, data, name=""):
    """

    :param data:
    :param name:
    """
    for i, rec in enumerate(data):
      rec['color'] = self._report.theme.colors[i+1]
      self.set_rcolors(rec['color'], rec['children'])
    self._datasets = [{'name': name, 'children': data, 'color': self._report.theme.colors[0]}]
    return self

  @property
  def _js__convertor__(self):
    return '''
      var result = [{name: options.x_column, children: []}]; var sizeTree = options.y_columns.length-1;
      data.forEach(function(rec){
        var path = []; var tmpResultLevel = result[0].children; var branchVal = 0;
        options.y_columns.forEach(function(s, i){
          var treeLevel = -1; 
          tmpResultLevel.forEach(function(l, j){if(l.name == rec[s]){treeLevel = j}});
          if(i == sizeTree){
            if(treeLevel >= 0){
              tmpResultLevel[treeLevel].size += rec[options.x_column]}else{tmpResultLevel.push({name: rec[s], size: rec[options.x_column]})}
          }else{
            if(treeLevel < 0 ){
              tmpResultLevel.push({name: rec[s], children: []}); treeLevel = tmpResultLevel.length - 1};
              tmpResultLevel = tmpResultLevel[treeLevel].children}
        })})'''


class ChartBoxPlot(Chart):

  @property
  def dom(self):
    """
    :rtype: JsNvd3.JsNvd3BoxPlot
    """
    if self._dom is None:
      self._dom = JsNvd3.JsNvd3BoxPlot(self._report, varName=self.chartId)
    return self._dom

  def add_box(self, q1, q3=None, outliers=None, maxRegularValue=None, mean=None, median=None, minRegularValue=None,
              minOutlier=None, maxOutlier=None, title=None):
    """

    https://github.com/nvd3-community/nvd3/blob/gh-pages/examples/boxPlotCustomModel.html

    :param q1:
    :param q3:
    :param outliers:
    :param maxRegularValue:
    :param mean:
    :param median:
    :param minRegularValue:
    :param minOutlier:
    :param maxOutlier:
    """
    names = ['q1', 'median', 'q3', 'outlData', 'maxRegularValue', 'mean', 'minRegularValue', 'minOutlier', 'maxOutlier']
    row = {}
    for i, val in enumerate([q1, median, q3, outliers, maxRegularValue, mean, minRegularValue, minOutlier, maxOutlier]):
      if val is not None:
        row[names[i]] = val
      elif names[i] == 'outlData':
        row['outlData'] = []
    series_id = len(self._datasets) - 1
    row['seriesColor'] = self._report.theme.colors[series_id]
    row['title'] = title or "Series %s" % series_id
    self._datasets.append(row)
    return self

  def add_trace(self, data, name=""):
    """

    :param data:
    :param name:
    """
    self._datasets = data
    return self


class ChartCandlestick(Chart):

  @property
  def dom(self):
    """
    :rtype: JsNvd3.JsNvd3BoxPlot
    """
    if self._dom is None:
      self._dom = JsNvd3.JsNvd3CandlestickBar(self._report, varName=self.chartId)
    return self._dom


class ChartOhlcBar(Chart):

  @property
  def dom(self):
    """
    :rtype: JsNvd3.JsNvd3OhlcBar
    """
    if self._dom is None:
      self._dom = JsNvd3.JsNvd3OhlcBar(self._report, varName=self.chartId)
    return self._dom


class ChartForceDirected(Chart):

  @property
  def dom(self):
    """
    :rtype: JsNvd3.JsNvd3BoxPlot
    """
    if self._dom is None:
      self._dom = JsNvd3.JsNvd3ForceDirectedGraph(self._report, varName=self.chartId)
    return self._dom

  def add_trace(self, data, name=""):
    """

    :param data:
    :param name:
    """
    for d in data.get('nodes', []):
      d['color'] = self._report.theme.colors[d.get('group', 1)]
    self._datasets = data
    return self



"""
CSS Style module for the scroll bar in the page

Track
::-webkit-scrollbar
::-webkit-scrollbar-track

Thumb
::-webkit-scrollbar-thumb
::-webkit-scrollbar-thumb:hover
::-webkit-scrollbar-thumb:active
::-webkit-scrollbar-thumb:vertical
::-webkit-scrollbar-thumb:horizontal

Buttons
Up
::-webkit-scrollbar-button
::-webkit-scrollbar-button:vertical:decrement
::-webkit-scrollbar-button:vertical:decrement:hover

Down
::-webkit-scrollbar-button:vertical:increment
::-webkit-scrollbar-button:vertical:increment:hover

Left
::-webkit-scrollbar-button:horizontal:decrement
::-webkit-scrollbar-button:horizontal:decrement:hover

Right
::-webkit-scrollbar-button:horizontal:increment
::-webkit-scrollbar-button:horizontal:increment:hover
"""

from epyk.core.css.styles.classes import CssStyle


class CssScrollBar(CssStyle.Style):
  _attrs = {"width": "14px", "height": "14px"}
  _selector = '::-webkit-scrollbar'


class CssScrollBarTrack(CssStyle.Style):
  _attrs = {"background": "#303030", "border": "solid 2px rgba(236, 0,0 , 0.5)"}
  _selector = '::-webkit-scrollbar-track'


class CssScrollBarTrackThumb(CssStyle.Style):
  _attrs = {"background": "#ffcf17"}
  _hover = {"background": "green"}
  _active = {"background": "green"}

  _selector = "::-webkit-scrollbar-thumb"


# class CssScrollBarTrackThumbHorizontal(CssStyle.Style):
#   attrs = {
#     "border-right": "solid 2px rgba(33,33,33,0.5)",
#     "border-left": "solid 2px rgba(33,33,33,0.5)",
#            }
#   cssId = {'reference': '::-webkit-scrollbar-thumb:horizontal'}

